from __future__ import annotations

from latentscore import FirstAudioSpinner, Streamable
from latentscore.spinner import Spinner


def test_spinner_disabled_is_noop() -> None:
    spinner = Spinner("Downloading", enabled=False)
    spinner.start()
    spinner.update("Still downloading")
    spinner.stop()


def test_first_audio_spinner_hooks_noop_when_disabled() -> None:
    spinner = FirstAudioSpinner(enabled=False)
    hooks = spinner.hooks()
    assert hooks.on_stream_start
    hooks.on_stream_start()
    assert hooks.on_item_preview_start
    hooks.on_item_preview_start(
        0,
        Streamable(content="vibe", duration=0.01),
        "embedding",
    )
    assert hooks.on_first_audio_chunk
    hooks.on_first_audio_chunk()
    assert hooks.on_stream_end
    hooks.on_stream_end()



from __future__ import annotations


class LatentScoreError(Exception):
    """Base error for the LatentScore library."""


class InvalidConfigError(LatentScoreError):
    """Raised when a config cannot be parsed or validated."""


class LLMInferenceError(LatentScoreError):
    """Raised when a model provider fails to produce a response."""


class ConfigGenerateError(LatentScoreError):
    """Raised when a model fails while generating a config."""


class ModelNotAvailableError(LatentScoreError):
    """Raised when required model weights or dependencies are missing."""


class PlaybackError(LatentScoreError):
    """Raised when audio playback is unavailable or fails."""



from __future__ import annotations

import asyncio
import json
import logging
import warnings
from collections.abc import Mapping
from threading import Event, Lock, Thread
from typing import Any, Callable, Coroutine, Literal, cast

from pydantic import BaseModel, ValidationError

from ..config import MusicConfig, MusicConfigPromptPayload
from ..errors import ConfigGenerateError, LLMInferenceError, ModelNotAvailableError
from ..models import EXTERNAL_PREFIX, build_litellm_prompt

_DEFAULT_TEMPERATURE = 0.0
_atexit_guard_registered = False
_safe_get_event_loop_installed = False
_LOGGER = logging.getLogger("latentscore.providers.litellm")
_RESERVED_LITELLM_KWARGS = frozenset(
    {"model", "messages", "response_format", "api_key", "reasoning_effort"}
)
ReasoningEffort = Literal["none", "minimal", "low", "medium", "high", "xhigh", "default"]
_litellm_loop: asyncio.AbstractEventLoop | None = None
_litellm_thread: Thread | None = None
_litellm_ready = Event()
_litellm_lock = Lock()
_litellm_shutdown_registered = False
_litellm_logging_configured = False

try:
    from json_repair import repair_json  # type: ignore[import]  # Optional dependency.
except ImportError:  # pragma: no cover - optional dependency
    repair_json: Callable[[str], str] | None = None


def _safe_async_cleanup(
    cleanup_coro: Callable[[], Coroutine[Any, Any, None]],
) -> None:
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        _LOGGER.debug("LiteLLM cleanup running outside an event loop.")
        loop = None

    if loop is not None:
        try:
            loop.create_task(cleanup_coro())
        except Exception as exc:
            _LOGGER.warning("LiteLLM async cleanup task failed: %s", exc, exc_info=True)
        return

    try:
        asyncio.run(cleanup_coro())
    except Exception as exc:
        _LOGGER.warning("LiteLLM async cleanup failed: %s", exc, exc_info=True)


def _install_safe_get_event_loop() -> None:
    global _safe_get_event_loop_installed
    if _safe_get_event_loop_installed:
        return
    _safe_get_event_loop_installed = True

    original_get_event_loop = asyncio.get_event_loop

    def safe_get_event_loop() -> asyncio.AbstractEventLoop:
        try:
            loop = original_get_event_loop()
        except RuntimeError:
            _LOGGER.debug("LiteLLM cleanup creating a new event loop.")
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            return loop
        if loop.is_closed():
            _LOGGER.debug("LiteLLM cleanup replacing a closed event loop.")
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        return loop

    asyncio.get_event_loop = safe_get_event_loop  # type: ignore[assignment]

    try:
        from litellm.llms.custom_httpx.async_client_cleanup import (
            close_litellm_async_clients,
        )
    except Exception as exc:
        _LOGGER.debug("LiteLLM cleanup import unavailable: %s", exc)
        return
    _safe_async_cleanup(close_litellm_async_clients)


def _register_safe_get_event_loop_atexit() -> None:
    global _atexit_guard_registered
    if _atexit_guard_registered:
        return
    _atexit_guard_registered = True
    import atexit

    atexit.register(_install_safe_get_event_loop)


def _shutdown_litellm_loop() -> None:
    loop = _litellm_loop
    thread = _litellm_thread
    if loop is None or thread is None:
        return
    if loop.is_closed():
        return
    try:
        loop.call_soon_threadsafe(loop.stop)
    except Exception as exc:
        _LOGGER.info("LiteLLM loop stop failed: %s", exc, exc_info=True)
    thread.join(timeout=1.0)


def _register_litellm_loop_atexit() -> None:
    global _litellm_shutdown_registered
    if _litellm_shutdown_registered:
        return
    _litellm_shutdown_registered = True
    import atexit

    atexit.register(_shutdown_litellm_loop)


def _configure_litellm_logging(litellm_module: Any) -> None:
    global _litellm_logging_configured
    if _litellm_logging_configured:
        return
    _litellm_logging_configured = True
    try:
        litellm_module.turn_off_message_logging = True
        litellm_module.disable_streaming_logging = True
        litellm_module.logging = False
    except Exception as exc:
        _LOGGER.info("LiteLLM logging config failed: %s", exc, exc_info=True)
    warnings.filterwarnings("ignore", message="Pydantic serializer warnings")


def _ensure_litellm_loop() -> asyncio.AbstractEventLoop:
    global _litellm_loop, _litellm_thread
    with _litellm_lock:
        if _litellm_loop is not None and not _litellm_loop.is_closed():
            return _litellm_loop
        _litellm_ready.clear()

        def _runner() -> None:
            global _litellm_loop
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            _litellm_loop = loop
            _litellm_ready.set()
            loop.run_forever()
            try:
                loop.run_until_complete(loop.shutdown_asyncgens())
            except Exception as exc:
                _LOGGER.info("LiteLLM loop shutdown failed: %s", exc, exc_info=True)
            loop.close()

        _litellm_thread = Thread(
            target=_runner,
            name="latentscore-litellm-loop",
            daemon=True,
        )
        _litellm_thread.start()
    _litellm_ready.wait()
    if _litellm_loop is None:
        raise RuntimeError("LiteLLM loop failed to start")
    _register_litellm_loop_atexit()
    return _litellm_loop


async def _run_on_litellm_loop(coro: Coroutine[Any, Any, Any]) -> Any:
    loop = _ensure_litellm_loop()
    try:
        running = asyncio.get_running_loop()
    except RuntimeError:
        running = None
    if running is loop:
        return await coro
    future = asyncio.run_coroutine_threadsafe(coro, loop)
    try:
        return await asyncio.wrap_future(future)
    except asyncio.CancelledError:
        future.cancel()
        raise


def _extract_json_payload(content: str) -> str | None:
    start = content.find("{")
    end = content.rfind("}")
    if start == -1 or end == -1 or end <= start:
        return None
    return content[start : end + 1]


def _extract_config_payload(payload: str) -> Mapping[str, Any] | None:
    try:
        data = json.loads(payload)
    except json.JSONDecodeError:
        return None
    match data:
        case {"config": config} if isinstance(config, dict):
            assert isinstance(config, dict)
            return cast(dict[str, Any], config)
        case _:
            return None


def _content_snippet(content: str, limit: int = 200) -> str:
    cleaned = content.strip()
    if len(cleaned) <= limit:
        return cleaned
    return f"{cleaned[:limit]}..."


class _LiteLLMRequest(BaseModel):
    model: str
    messages: list[dict[str, str]]
    temperature: float | None = None
    response_format: type[BaseModel] | None = None
    api_key: str | None = None


class LiteLLMAdapter:
    """LiteLLM model wrapper implementing the async model protocol."""

    def __init__(
        self,
        model: str,
        *,
        api_key: str | None = None,
        litellm_kwargs: Mapping[str, Any] | None = None,
        temperature: float | None = None,
        system_prompt: str | None = None,
        response_format: type[BaseModel] | None = MusicConfigPromptPayload,
        reasoning_effort: ReasoningEffort | None = None,
    ) -> None:
        if model.startswith(EXTERNAL_PREFIX):
            model = model.removeprefix(EXTERNAL_PREFIX)
        self._model = model
        self._api_key = api_key
        self._litellm_kwargs = dict(litellm_kwargs or {})
        self._temperature = temperature
        self._system_prompt = system_prompt
        self._base_prompt = build_litellm_prompt()
        self._response_format: type[BaseModel] | None = response_format
        self._reasoning_effort: ReasoningEffort | None = reasoning_effort
        # self._schema_str = ""
        # if self._response_format:
        #     self._schema_str = f"<schema>{json.dumps(self._response_format.model_json_schema(), separators=(',', ':'))}</schema>"
        if self._api_key is None:
            _LOGGER.debug("No API key provided; letting LiteLLM read from env vars.")
        invalid_keys = _RESERVED_LITELLM_KWARGS.intersection(self._litellm_kwargs)
        if invalid_keys:
            keys = ", ".join(sorted(invalid_keys))
            raise ConfigGenerateError(f"litellm_kwargs cannot override: {keys}")

    def close(self) -> None:
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            _LOGGER.info("LiteLLM close running outside event loop.")
            asyncio.run(self.aclose())
            return
        loop.create_task(self.aclose())

    async def aclose(self) -> None:
        try:
            import litellm  # type: ignore[import]  # Optional dependency at runtime.
        except ImportError as exc:
            _LOGGER.info("LiteLLM not installed; skipping async close: %s", exc)
            return

        async def _close() -> None:
            _register_safe_get_event_loop_atexit()
            close_fn: Any = getattr(litellm, "aclose", None)
            if close_fn is None:
                close_fn = getattr(litellm, "close_litellm_async_clients", None)
            if close_fn is None:
                return
            await close_fn()

        try:
            await _run_on_litellm_loop(_close())
        except Exception as exc:
            _LOGGER.warning("LiteLLM close failed: %s", exc, exc_info=True)

    async def generate(self, vibe: str) -> MusicConfig:
        try:
            import litellm  # type: ignore[import]
            from litellm import acompletion  # type: ignore[import]
        except ImportError as exc:
            _LOGGER.warning("LiteLLM not installed: %s", exc)
            raise ModelNotAvailableError("litellm is not installed") from exc

        _register_safe_get_event_loop_atexit()
        _configure_litellm_logging(litellm)
        messages: list[dict[str, str]] = [{"role": "system", "content": self._base_prompt}]
        if self._system_prompt:
            messages.append({"role": "system", "content": self._system_prompt})

        messages.append(
            {
                "role": "user",
                # "content": f"{self._schema_str}\n<vibe>{vibe}</vibe>\n<output>",
                "content": f"<vibe>{vibe}</vibe>\n<output>",
            }
        )

        request = _LiteLLMRequest(
            model=self._model,
            messages=messages,
            temperature=self._temperature,
            response_format=self._response_format,
            api_key=self._api_key or None,
        ).model_dump(exclude_none=True)
        request.update(self._litellm_kwargs)

        try:
            response: Any = await _run_on_litellm_loop(
                acompletion(**request, reasoning_effort=self._reasoning_effort)
            )
        except Exception as exc:  # pragma: no cover - provider errors
            _LOGGER.warning("LiteLLM request failed: %s", exc, exc_info=True)
            raise LLMInferenceError(str(exc)) from exc

        content = ""
        try:
            assert response is not None
            if not hasattr(response, "choices"):
                raise ConfigGenerateError("LiteLLM response missing choices")
            raw_content = response.choices[0].message.content
            if not isinstance(raw_content, str) or not raw_content.strip():
                raise ConfigGenerateError("LiteLLM returned empty content")
            content = raw_content.strip()
            print(self._parse_config_payload(content))
            return self._parse_config_payload(content)
        except ValidationError as exc:
            extracted = _extract_json_payload(content)
            if extracted:
                try:
                    return self._parse_config_payload(extracted)
                except ValidationError:
                    _LOGGER.warning(
                        "LiteLLM returned invalid JSON after extraction.", exc_info=True
                    )
            if repair_json is not None:
                repaired = repair_json(content)
                try:
                    return self._parse_config_payload(repaired)
                except ValidationError:
                    _LOGGER.warning("LiteLLM returned invalid JSON after repair.", exc_info=True)
            snippet = _content_snippet(content) or "<empty>"
            _LOGGER.warning("LiteLLM returned invalid JSON: %s", snippet)
            raise ConfigGenerateError(f"LiteLLM returned non-JSON content: {snippet}") from exc

    def _parse_config_payload(self, payload: str) -> MusicConfig:
        if self._response_format is MusicConfigPromptPayload:
            try:
                wrapper = MusicConfigPromptPayload.model_validate_json(payload)
            except ValidationError:
                config_payload = _extract_config_payload(payload)
                if config_payload is not None:
                    return MusicConfig.model_validate(config_payload)
                return MusicConfig.model_validate_json(payload)
            return wrapper.config.to_config()
        return MusicConfig.model_validate_json(payload)



# pyright: reportPrivateUsage=false

from __future__ import annotations

import asyncio
import inspect
import logging
import os
from collections import deque
from collections.abc import AsyncIterable, AsyncIterator, Iterable, Iterator
from concurrent.futures import ThreadPoolExecutor
from queue import Queue
from threading import Thread, Timer
from typing import Any, Callable, Coroutine, Literal, TypeGuard, TypeVar

import numpy as np
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr, field_validator

from .audio import SAMPLE_RATE, AudioNumbers, FloatArray, ensure_audio_contract, write_wav
from .config import (
    ConfigInput,
    MusicConfig,
    MusicConfigUpdate,
    SynthConfig,
    UpdateInput,
    _MusicConfigInternal,
    _MusicConfigUpdateInternal,
    coerce_internal_config,
    coerce_internal_update,
    is_empty_update,
    merge_internal_config,
)
from .errors import InvalidConfigError, ModelNotAvailableError
from .logging_utils import log_exception
from .models import (
    EXTERNAL_PREFIX,
    ExternalModelSpec,
    ModelForGeneratingMusicConfig,
    ModelSpec,
    resolve_model,
)
from .spinner import Spinner, render_error
from .synth import assemble, interpolate_configs

StreamContent = str | ConfigInput | UpdateInput
FallbackPolicy = Literal["none", "keep_last", "embedding"]
FallbackInput = FallbackPolicy | ConfigInput | UpdateInput
ModelLoadRole = Literal["primary", "fallback"]


class StreamEvent(BaseModel):
    kind: Literal[
        "model_load_start",
        "model_load_end",
        "stream_start",
        "item_resolve_start",
        "item_resolve_success",
        "item_resolve_error",
        "item_preview_start",
        "first_config_ready",
        "first_audio_chunk",
        "stream_end",
        "fallback_used",
        "error",
    ]
    model: object | None = None
    model_role: ModelLoadRole | None = None
    index: int | None = None
    item: Streamable | None = None
    error: Exception | None = None
    fallback: FallbackInput | None = None
    preview: bool | None = None

    model_config = ConfigDict(
        frozen=True,
        extra="forbid",
        arbitrary_types_allowed=True,
    )

    @field_validator("model")
    @classmethod
    def _validate_model(cls, value: object | None) -> object | None:
        if value is None:
            return None
        if isinstance(value, str):
            return value
        if isinstance(value, ExternalModelSpec):
            return value
        if hasattr(value, "generate"):
            return value
        raise InvalidConfigError("model must be a ModelSpec-compatible value")


class StreamHooks(BaseModel):
    on_event: Callable[[StreamEvent], None] | None = None
    on_model_load_start: Callable[[ModelSpec, ModelLoadRole], None] | None = None
    on_model_load_end: Callable[[ModelSpec, ModelLoadRole], None] | None = None
    on_stream_start: Callable[[], None] | None = None
    on_item_resolve_start: Callable[[int, Streamable], None] | None = None
    on_item_resolve_success: Callable[[int, Streamable], None] | None = None
    on_item_resolve_error: (
        Callable[[int, Streamable, Exception, FallbackInput | None], None] | None
    ) = None
    on_item_preview_start: Callable[[int, Streamable, bool], None] | None = None
    on_first_config_ready: Callable[[int, Streamable], None] | None = None
    on_first_audio_chunk: Callable[[], None] | None = None
    on_stream_end: Callable[[], None] | None = None
    on_error: Callable[[Exception], None] | None = None

    model_config = ConfigDict(
        frozen=True,
        extra="forbid",
        arbitrary_types_allowed=True,
    )


class RenderHooks(BaseModel):
    on_start: Callable[[], None] | None = None
    on_model_start: Callable[[ModelSpec], None] | None = None
    on_model_end: Callable[[ModelSpec], None] | None = None
    on_synth_start: Callable[[], None] | None = None
    on_synth_end: Callable[[], None] | None = None
    on_end: Callable[[], None] | None = None
    on_error: Callable[[Exception], None] | None = None

    model_config = ConfigDict(
        frozen=True,
        extra="forbid",
        arbitrary_types_allowed=True,
    )


class FirstAudioSpinner:
    """Spinner helper for first-audio latency with optional preview messaging."""

    def __init__(
        self,
        *,
        delay: float = 0.35,
        enabled: bool | None = None,
    ) -> None:
        self._delay = delay
        self._spinner = Spinner("Generating first audio", enabled=enabled)
        self._timer: Timer | None = None
        self._started = False
        self._stopped = False

    def hooks(self) -> StreamHooks:
        return StreamHooks(
            on_model_load_start=self._on_model_load_start,
            on_model_load_end=self._on_model_load_end,
            on_stream_start=self._on_stream_start,
            on_item_preview_start=self._on_preview_start,
            on_first_audio_chunk=self._on_first_audio_chunk,
            on_stream_end=self._on_stream_end,
        )

    def _on_model_load_start(self, model: ModelSpec, role: ModelLoadRole) -> None:
        _ = model
        if role != "primary":
            return
        if self._stopped:
            return
        self._spinner.update("Loading model")
        self._start_immediately()

    def _on_model_load_end(self, model: ModelSpec, role: ModelLoadRole) -> None:
        _ = model
        if role != "primary":
            return
        if self._stopped:
            return
        self._spinner.update("Generating first audio")

    def _on_stream_start(self) -> None:
        if self._stopped or self._started or self._timer is not None:
            return
        self._timer = Timer(self._delay, self._start)
        self._timer.daemon = True
        self._timer.start()

    def _on_preview_start(
        self,
        index: int,
        item: Streamable,
        preview: bool,
    ) -> None:
        _ = index
        _ = item
        if preview:
            self._spinner.update(
                "Generating first audio (speculative preview running; set preview=False to disable)"
            )

    def _on_first_audio_chunk(self) -> None:
        self._stop()

    def _on_stream_end(self) -> None:
        self._stop()

    def _start(self) -> None:
        if self._stopped or self._started:
            return
        self._started = True
        self._spinner.start()

    def _start_immediately(self) -> None:
        if self._stopped or self._started:
            return
        if self._timer is not None:
            self._timer.cancel()
            self._timer = None
        self._start()

    def _stop(self) -> None:
        if self._stopped:
            return
        self._stopped = True
        if self._timer is not None:
            self._timer.cancel()
            self._timer = None
        self._spinner.stop()


_MIN_CHUNK_SECONDS = 1e-6
_MIN_RECOMMENDED_DURATION_SECONDS = 5.0
_MIN_RECOMMENDED_TRANSITION_SECONDS = 1.0
_LOGGER = logging.getLogger("latentscore.main")
_EXECUTOR = ThreadPoolExecutor(max_workers=1)


class _PendingTaskError(Exception):
    def __init__(self, error: Exception) -> None:
        super().__init__(str(error))
        self.error = error


def _suppress_task_exception(task: asyncio.Task[object]) -> None:
    try:
        _ = task.exception()
    except asyncio.CancelledError:
        return
    except Exception as exc:
        _LOGGER.debug("Background task failed: %s", exc, exc_info=True)


def _cancel_task(task: asyncio.Task[object]) -> None:
    if task.done():
        _suppress_task_exception(task)
        return
    task.add_done_callback(_suppress_task_exception)
    task.cancel()


class Streamable(BaseModel):
    """Structured streaming input with timing controls."""

    content: StreamContent
    duration: float = Field(default=60.0, gt=0.0)
    transition_duration: float = Field(default=1.0, ge=0.0)
    fallback: FallbackInput | None = None
    preview: bool | None = None

    model_config = ConfigDict(extra="forbid")


T = TypeVar("T")


def _run_async(coro: Coroutine[Any, Any, T]) -> T:
    try:
        asyncio.get_running_loop()
    except RuntimeError:
        _LOGGER.debug("No running event loop; running coroutine directly.")
        return asyncio.run(coro)

    return _EXECUTOR.submit(lambda: asyncio.run(coro)).result()


async def _maybe_close_model(model: ModelForGeneratingMusicConfig | None) -> None:
    if model is None:
        return
    aclose = getattr(model, "aclose", None)
    if callable(aclose):
        try:
            result = aclose()
            if inspect.isawaitable(result):
                await result
            return
        except Exception as exc:
            _LOGGER.warning("Model async close failed: %s", exc, exc_info=True)
            return
    close = getattr(model, "close", None)
    if callable(close):
        try:
            close()
        except Exception as exc:
            _LOGGER.warning("Model close failed: %s", exc, exc_info=True)


def _should_auto_close_model(model: ModelSpec) -> bool:
    if isinstance(model, ExternalModelSpec):
        return True
    if isinstance(model, str):
        return model.startswith(EXTERNAL_PREFIX)
    return False


def _log_exception(context: str, exc: Exception, *, show_console: bool = True) -> None:
    debug = bool(os.environ.get("LATENTSCORE_DEBUG"))
    _LOGGER.warning("%s failed: %s", context, exc, exc_info=debug)
    log_exception(context, exc)
    if show_console:
        render_error(context, exc)


def _is_model_spec(value: object) -> TypeGuard[ModelSpec]:
    if isinstance(value, str):
        return True
    if isinstance(value, ExternalModelSpec):
        return True
    return hasattr(value, "generate")


def _emit_event(hooks: StreamHooks | None, event: StreamEvent) -> None:
    if hooks is None:
        return
    try:
        if hooks.on_event is not None:
            hooks.on_event(event)
        match event.kind:
            case "model_load_start":
                if (
                    hooks.on_model_load_start is not None
                    and event.model is not None
                    and event.model_role is not None
                ):
                    if _is_model_spec(event.model):
                        hooks.on_model_load_start(event.model, event.model_role)
            case "model_load_end":
                if (
                    hooks.on_model_load_end is not None
                    and event.model is not None
                    and event.model_role is not None
                ):
                    if _is_model_spec(event.model):
                        hooks.on_model_load_end(event.model, event.model_role)
            case "stream_start":
                if hooks.on_stream_start is not None:
                    hooks.on_stream_start()
            case "item_resolve_start":
                if (
                    hooks.on_item_resolve_start is not None
                    and event.index is not None
                    and event.item is not None
                ):
                    hooks.on_item_resolve_start(event.index, event.item)
            case "item_resolve_success":
                if (
                    hooks.on_item_resolve_success is not None
                    and event.index is not None
                    and event.item is not None
                ):
                    hooks.on_item_resolve_success(event.index, event.item)
            case "item_resolve_error":
                if (
                    hooks.on_item_resolve_error is not None
                    and event.index is not None
                    and event.item is not None
                    and event.error is not None
                ):
                    hooks.on_item_resolve_error(
                        event.index,
                        event.item,
                        event.error,
                        event.fallback,
                    )
            case "item_preview_start":
                if (
                    hooks.on_item_preview_start is not None
                    and event.index is not None
                    and event.item is not None
                    and event.preview is not None
                ):
                    hooks.on_item_preview_start(
                        event.index,
                        event.item,
                        event.preview,
                    )
            case "first_config_ready":
                if (
                    hooks.on_first_config_ready is not None
                    and event.index is not None
                    and event.item is not None
                ):
                    hooks.on_first_config_ready(event.index, event.item)
            case "first_audio_chunk":
                if hooks.on_first_audio_chunk is not None:
                    hooks.on_first_audio_chunk()
            case "stream_end":
                if hooks.on_stream_end is not None:
                    hooks.on_stream_end()
            case "error":
                if hooks.on_error is not None and event.error is not None:
                    hooks.on_error(event.error)
            case "fallback_used":
                pass
            case _:
                pass
    except Exception as exc:
        debug = bool(os.environ.get("LATENTSCORE_DEBUG"))
        _LOGGER.warning("Stream hook failed: %s", exc, exc_info=debug)


def _emit_render(
    hooks: RenderHooks | None,
    *,
    kind: Literal[
        "start",
        "model_start",
        "model_end",
        "synth_start",
        "synth_end",
        "end",
        "error",
    ],
    model: ModelSpec | None = None,
    error: Exception | None = None,
) -> None:
    if hooks is None:
        return
    try:
        match kind:
            case "start":
                if hooks.on_start is not None:
                    hooks.on_start()
            case "model_start":
                if hooks.on_model_start is not None and model is not None:
                    hooks.on_model_start(model)
            case "model_end":
                if hooks.on_model_end is not None and model is not None:
                    hooks.on_model_end(model)
            case "synth_start":
                if hooks.on_synth_start is not None:
                    hooks.on_synth_start()
            case "synth_end":
                if hooks.on_synth_end is not None:
                    hooks.on_synth_end()
            case "end":
                if hooks.on_end is not None:
                    hooks.on_end()
            case "error":
                if hooks.on_error is not None and error is not None:
                    hooks.on_error(error)
            case _:
                raise InvalidConfigError(f"Unknown render hook kind: {kind}")
    except Exception as exc:
        debug = bool(os.environ.get("LATENTSCORE_DEBUG"))
        _LOGGER.warning("Render hook failed: %s", exc, exc_info=debug)


class _PrefetchedItem(BaseModel):
    index: int
    item: Streamable
    task: asyncio.Task[_MusicConfigInternal] | None

    model_config = ConfigDict(
        extra="forbid",
        arbitrary_types_allowed=True,
    )


class _ModelLoadState(BaseModel):
    model: Any
    role: ModelLoadRole
    hooks: StreamHooks | None
    loaded: bool = False
    _load_task: asyncio.Task[None] | None = PrivateAttr(default=None)

    model_config = ConfigDict(
        extra="forbid",
        arbitrary_types_allowed=True,
    )

    @field_validator("model")
    @classmethod
    def _validate_model(cls, value: object) -> object:
        if not hasattr(value, "generate"):
            raise InvalidConfigError("model must implement generate()")
        return value

    async def ensure_loaded(self) -> None:
        if self.loaded:
            return
        if self._load_task is None:
            self._load_task = asyncio.create_task(self._run_load())
        await self._load_task

    async def _run_load(self) -> None:
        if self.loaded:
            return
        self.loaded = True
        if self.hooks is not None:
            _emit_event(
                self.hooks,
                StreamEvent(
                    kind="model_load_start",
                    model=self.model,
                    model_role=self.role,
                ),
            )
        try:
            warmup = getattr(self.model, "warmup", None)
            if callable(warmup):
                await asyncio.to_thread(warmup)
        finally:
            if self.hooks is not None:
                _emit_event(
                    self.hooks,
                    StreamEvent(
                        kind="model_load_end",
                        model=self.model,
                        model_role=self.role,
                    ),
                )


def _to_synth_config(config: _MusicConfigInternal) -> SynthConfig:
    return config


def _from_synth_config(config: SynthConfig) -> _MusicConfigInternal:
    return config


def _default_internal_config() -> _MusicConfigInternal:
    return MusicConfig().to_internal()


def _apply_update(
    base: _MusicConfigInternal,
    update: UpdateInput | None,
) -> _MusicConfigInternal:
    match update:
        case None:
            return base
        case _:
            internal_update = coerce_internal_update(update)
            return merge_internal_config(base, internal_update)


def _chunk_count(duration: float, chunk_seconds: float) -> int:
    return max(1, int(round(duration / max(chunk_seconds, _MIN_CHUNK_SECONDS))))


def _transition_steps(transition_duration: float, chunk_seconds: float) -> int:
    match transition_duration:
        case (float() | int()) as duration if duration <= 0:
            return 0
        case (float() | int()) as duration:
            return max(1, int(round(duration / max(chunk_seconds, _MIN_CHUNK_SECONDS))))
        case _:
            raise InvalidConfigError("transition_duration must be a number")


def _iter_transition_configs(
    current: _MusicConfigInternal | None,
    target: _MusicConfigInternal,
    steps: int,
) -> Iterator[_MusicConfigInternal]:
    if current is None or steps <= 1:
        yield target
        return
    start = _to_synth_config(current)
    end = _to_synth_config(target)
    for step in range(1, steps + 1):
        t = step / steps
        interpolated = interpolate_configs(start, end, t)
        yield _from_synth_config(interpolated)


def _warn_aggressive_timing(
    *,
    duration: float,
    transition_duration: float,
    warned: dict[str, bool],
) -> None:
    if duration < _MIN_RECOMMENDED_DURATION_SECONDS and not warned["duration"]:
        warned["duration"] = True
        _LOGGER.warning(
            "Track duration %.2fs is very short (<%.1fs). "
            "This can reduce quality, and the model may not generate configs fast enough.",
            duration,
            _MIN_RECOMMENDED_DURATION_SECONDS,
        )
    if transition_duration < _MIN_RECOMMENDED_TRANSITION_SECONDS and not warned["transition"]:
        warned["transition"] = True
        _LOGGER.warning(
            "Transition duration %.2fs is very short (<%.1fs). "
            "Rapid transitions can significantly impact quality.",
            transition_duration,
            _MIN_RECOMMENDED_TRANSITION_SECONDS,
        )


def _wrap_streamables(
    items: Iterable[StreamContent],
    *,
    duration: float,
    transition_duration: float,
) -> Iterator[Streamable]:
    for item in items:
        yield Streamable(
            content=item,
            duration=duration,
            transition_duration=transition_duration,
        )


def render(
    vibe: str,
    *,
    duration: float = 8.0,
    model: ModelSpec = "fast",
    config: ConfigInput | None = None,
    update: UpdateInput | None = None,
    hooks: RenderHooks | None = None,
) -> FloatArray:
    _emit_render(hooks, kind="start")
    try:
        if config is None:
            resolved = resolve_model(model)
            should_close = _should_auto_close_model(model)
            _emit_render(hooks, kind="model_start", model=model)

            async def _generate_with_cleanup() -> MusicConfig:
                try:
                    return await resolved.generate(vibe)
                finally:
                    if should_close:
                        await _maybe_close_model(resolved)

            try:
                base = _run_async(_generate_with_cleanup()).to_internal()
            finally:
                _emit_render(hooks, kind="model_end", model=model)
        else:
            base = coerce_internal_config(config)
        target = _apply_update(base, update)

        _emit_render(hooks, kind="synth_start")
        audio = assemble(_to_synth_config(target), duration)
        normalized = ensure_audio_contract(audio, sample_rate=SAMPLE_RATE)
        _emit_render(hooks, kind="synth_end")
        _emit_render(hooks, kind="end")
        return normalized
    except Exception as exc:
        _emit_render(hooks, kind="error", error=exc)
        show_console = hooks is None or hooks.on_error is None
        _log_exception("render", exc, show_console=show_console)
        raise


def _bridge_async_stream(
    async_iter: AsyncIterable[FloatArray],
    *,
    queue_maxsize: int = 0,
) -> Iterator[FloatArray]:
    sentinel = object()
    queue: Queue[object] = Queue(maxsize=queue_maxsize) if queue_maxsize > 0 else Queue()

    def _runner() -> None:
        async def _run() -> None:
            try:
                async for chunk in async_iter:
                    queue.put(chunk)
            except BaseException as exc:
                queue.put(exc)
            finally:
                queue.put(sentinel)

        asyncio.run(_run())

    thread = Thread(target=_runner, daemon=True)
    thread.start()
    while True:
        item = queue.get()
        if item is sentinel:
            break
        if isinstance(item, BaseException):
            raise item
        if not isinstance(item, np.ndarray):
            raise InvalidConfigError("Async stream emitted non-audio chunk")
        yield item
    thread.join(timeout=0)


def stream(
    items: Iterable[Streamable],
    *,
    chunk_seconds: float = 1.0,
    model: ModelSpec = "fast",
    config: ConfigInput | None = None,
    update: UpdateInput | None = None,
    prefetch_depth: int = 1,
    preview: bool = False,
    fallback: FallbackInput | None = None,
    fallback_model: ModelSpec = "fast",
    hooks: StreamHooks | None = None,
    queue_maxsize: int = 0,
) -> Iterable[FloatArray]:
    try:
        items_obj: object = items
        match items_obj:
            case Streamable():  # type: ignore[reportUnnecessaryComparison]
                raise InvalidConfigError("stream expects an iterable of Streamable items")
            case str():  # type: ignore[reportUnnecessaryComparison]
                raise InvalidConfigError("stream expects an iterable of Streamable items")
            case _:
                pass

        if prefetch_depth < 0:
            raise InvalidConfigError("prefetch_depth must be >= 0")

        async_iter = astream(
            items,
            chunk_seconds=chunk_seconds,
            model=model,
            config=config,
            update=update,
            prefetch_depth=prefetch_depth,
            preview=preview,
            fallback=fallback,
            fallback_model=fallback_model,
            hooks=hooks,
        )
        for chunk in _bridge_async_stream(async_iter, queue_maxsize=queue_maxsize):
            yield chunk
    except Exception as exc:
        show_console = hooks is None or hooks.on_error is None
        _log_exception("stream", exc, show_console=show_console)
        raise


def stream_texts(
    prompts: Iterable[str],
    *,
    duration: float = 60.0,
    transition_duration: float = 1.0,
    chunk_seconds: float = 1.0,
    model: ModelSpec = "fast",
    config: ConfigInput | None = None,
    update: UpdateInput | None = None,
    prefetch_depth: int = 1,
    preview: bool = False,
    fallback: FallbackInput | None = None,
    fallback_model: ModelSpec = "fast",
    hooks: StreamHooks | None = None,
) -> Iterable[FloatArray]:
    try:
        match prompts:
            case str():
                raise InvalidConfigError("stream_texts expects an iterable of strings")
            case _:
                pass

        items = _wrap_streamables(
            prompts,
            duration=duration,
            transition_duration=transition_duration,
        )
        return stream(
            items,
            chunk_seconds=chunk_seconds,
            model=model,
            config=config,
            update=update,
            prefetch_depth=prefetch_depth,
            preview=preview,
            fallback=fallback,
            fallback_model=fallback_model,
            hooks=hooks,
        )
    except Exception as exc:
        show_console = hooks is None or hooks.on_error is None
        _log_exception("stream_texts", exc, show_console=show_console)
        raise


def stream_configs(
    configs: Iterable[ConfigInput],
    *,
    duration: float = 60.0,
    transition_duration: float = 1.0,
    chunk_seconds: float = 1.0,
    model: ModelSpec = "fast",
    config: ConfigInput | None = None,
    update: UpdateInput | None = None,
    prefetch_depth: int = 1,
    preview: bool = False,
    fallback: FallbackInput | None = None,
    fallback_model: ModelSpec = "fast",
    hooks: StreamHooks | None = None,
) -> Iterable[FloatArray]:
    try:
        configs_obj: object = configs
        match configs_obj:
            case MusicConfig() | _MusicConfigInternal():  # type: ignore[reportUnnecessaryComparison]
                raise InvalidConfigError("stream_configs expects an iterable of config objects")
            case _:
                pass

        items = _wrap_streamables(
            configs,
            duration=duration,
            transition_duration=transition_duration,
        )
        return stream(
            items,
            chunk_seconds=chunk_seconds,
            model=model,
            config=config,
            update=update,
            prefetch_depth=prefetch_depth,
            preview=preview,
            fallback=fallback,
            fallback_model=fallback_model,
            hooks=hooks,
        )
    except Exception as exc:
        show_console = hooks is None or hooks.on_error is None
        _log_exception("stream_configs", exc, show_console=show_console)
        raise


def stream_updates(
    updates: Iterable[UpdateInput],
    *,
    duration: float = 60.0,
    transition_duration: float = 1.0,
    chunk_seconds: float = 1.0,
    model: ModelSpec = "fast",
    config: ConfigInput | None = None,
    update: UpdateInput | None = None,
    prefetch_depth: int = 1,
    preview: bool = False,
    fallback: FallbackInput | None = None,
    fallback_model: ModelSpec = "fast",
    hooks: StreamHooks | None = None,
) -> Iterable[FloatArray]:
    try:
        updates_obj: object = updates
        match updates_obj:
            case MusicConfigUpdate() | _MusicConfigUpdateInternal():  # type: ignore[reportUnnecessaryComparison]
                raise InvalidConfigError("stream_updates expects an iterable of update objects")
            case _:
                pass

        items = _wrap_streamables(
            updates,
            duration=duration,
            transition_duration=transition_duration,
        )
        return stream(
            items,
            chunk_seconds=chunk_seconds,
            model=model,
            config=config,
            update=update,
            prefetch_depth=prefetch_depth,
            preview=preview,
            fallback=fallback,
            fallback_model=fallback_model,
            hooks=hooks,
        )
    except Exception as exc:
        show_console = hooks is None or hooks.on_error is None
        _log_exception("stream_updates", exc, show_console=show_console)
        raise


def save_wav(path: str, audio_or_chunks: AudioNumbers | Iterable[AudioNumbers]) -> str:
    try:
        written = write_wav(path, audio_or_chunks, sample_rate=SAMPLE_RATE)
        return str(written)
    except Exception as exc:
        _log_exception("save_wav", exc, show_console=True)
        raise


def _coerce_async_items(
    items: Iterable[Streamable] | AsyncIterable[Streamable],
) -> AsyncIterable[Streamable]:
    items_obj: object = items
    match items_obj:
        case Streamable():  # type: ignore[reportUnnecessaryComparison]
            raise InvalidConfigError("astream expects an iterable of Streamable items")
        case str():  # type: ignore[reportUnnecessaryComparison]
            raise InvalidConfigError("astream expects an iterable of Streamable items")
        case AsyncIterable() as async_items:
            return async_items
        case Iterable() as sync_items:
            pass
        case _:
            raise InvalidConfigError("astream expects an iterable of Streamable items")

    async def _iter() -> AsyncIterator[Streamable]:
        for item in sync_items:
            yield item

    return _iter()


async def _generate_internal(
    vibe: str,
    model: ModelForGeneratingMusicConfig,
) -> _MusicConfigInternal:
    return (await model.generate(vibe)).to_internal()


async def _generate_internal_with_load(
    vibe: str,
    model: ModelForGeneratingMusicConfig,
    load_state: _ModelLoadState,
) -> _MusicConfigInternal:
    await load_state.ensure_loaded()
    return await _generate_internal(vibe, model)


async def _resolve_preview(
    item: Streamable,
    *,
    preview: bool,
    fallback_model: ModelForGeneratingMusicConfig,
    update: UpdateInput | None,
    fallback_load: _ModelLoadState | None = None,
) -> _MusicConfigInternal | None:
    if not preview or not isinstance(item.content, str):
        return None
    if fallback_load is not None:
        try:
            await fallback_load.ensure_loaded()
        except ModelNotAvailableError as exc:
            _LOGGER.warning("Preview model unavailable: %s", exc, exc_info=True)
            return None
        except Exception as exc:
            _LOGGER.warning("Preview model warmup failed: %s", exc, exc_info=True)
            return None
    try:
        preview_config = await fallback_model.generate(item.content)
    except ModelNotAvailableError as exc:
        _LOGGER.warning("Preview generation unavailable: %s", exc, exc_info=True)
        return None
    except Exception as exc:
        _LOGGER.warning("Preview generation failed: %s", exc, exc_info=True)
        return None
    return _apply_update(preview_config.to_internal(), update)


async def _resolve_preview_or_primary(
    *,
    item: Streamable,
    index: int,
    pending_task: asyncio.Task[_MusicConfigInternal],
    preview: bool,
    fallback_model: ModelForGeneratingMusicConfig,
    update: UpdateInput | None,
    fallback_load: _ModelLoadState | None,
    hooks: StreamHooks | None,
) -> tuple[_MusicConfigInternal, asyncio.Task[_MusicConfigInternal] | None]:
    if pending_task.done():
        try:
            target = _apply_update(pending_task.result(), update)
        except Exception as exc:
            raise _PendingTaskError(exc) from exc
        _emit_event(hooks, StreamEvent(kind="item_resolve_success", index=index, item=item))
        return target, None

    if not preview or not isinstance(item.content, str):
        try:
            target = _apply_update(await pending_task, update)
        except Exception as exc:
            raise _PendingTaskError(exc) from exc
        _emit_event(hooks, StreamEvent(kind="item_resolve_success", index=index, item=item))
        return target, None

    preview_task = asyncio.create_task(
        _resolve_preview(
            item,
            preview=preview,
            fallback_model=fallback_model,
            update=update,
            fallback_load=fallback_load,
        )
    )
    _done, _pending = await asyncio.wait(
        {pending_task, preview_task},
        return_when=asyncio.FIRST_COMPLETED,
    )
    if pending_task.done():
        _cancel_task(preview_task)
        try:
            target = _apply_update(pending_task.result(), update)
        except Exception as exc:
            raise _PendingTaskError(exc) from exc
        _emit_event(hooks, StreamEvent(kind="item_resolve_success", index=index, item=item))
        return target, None

    try:
        preview_config = await preview_task
    except Exception as exc:
        _LOGGER.warning("Preview task failed: %s", exc, exc_info=True)
        preview_config = None
    if preview_config is None:
        try:
            target = _apply_update(await pending_task, update)
        except Exception as exc:
            raise _PendingTaskError(exc) from exc
        _emit_event(hooks, StreamEvent(kind="item_resolve_success", index=index, item=item))
        return target, None

    _emit_event(
        hooks,
        StreamEvent(kind="item_preview_start", index=index, item=item, preview=preview),
    )
    return preview_config, pending_task


async def _resolve_fallback(
    item: Streamable,
    *,
    current: _MusicConfigInternal | None,
    fallback: FallbackInput | None,
    fallback_model: ModelForGeneratingMusicConfig,
    update: UpdateInput | None,
    fallback_load: _ModelLoadState | None = None,
) -> _MusicConfigInternal | None:
    if fallback is None or fallback == "keep_last":
        base = current or _default_internal_config()
        return _apply_update(base, update)
    if fallback == "none":
        return None
    if fallback == "embedding":
        if not isinstance(item.content, str):
            base = current or _default_internal_config()
            return _apply_update(base, update)
        if fallback_load is not None:
            try:
                await fallback_load.ensure_loaded()
            except ModelNotAvailableError as exc:
                _LOGGER.warning("Fallback model unavailable: %s", exc, exc_info=True)
            except Exception as exc:
                _LOGGER.warning("Fallback model warmup failed: %s", exc, exc_info=True)
        try:
            target = await fallback_model.generate(item.content)
        except ModelNotAvailableError as exc:
            _LOGGER.warning("Fallback generation unavailable: %s", exc, exc_info=True)
            base = current or _default_internal_config()
            return _apply_update(base, update)
        except Exception as exc:
            _LOGGER.warning("Fallback generation failed: %s", exc, exc_info=True)
            base = current or _default_internal_config()
            return _apply_update(base, update)
        return _apply_update(target.to_internal(), update)
    if fallback_load is not None and isinstance(fallback, str):
        try:
            await fallback_load.ensure_loaded()
        except ModelNotAvailableError as exc:
            _LOGGER.warning("Fallback model unavailable: %s", exc, exc_info=True)
            base = current or _default_internal_config()
            return _apply_update(base, update)
        except Exception as exc:
            _LOGGER.warning("Fallback model warmup failed: %s", exc, exc_info=True)
            base = current or _default_internal_config()
            return _apply_update(base, update)
    try:
        return await _resolve_target_async(
            fallback,
            current=current,
            model=fallback_model,
            update=update,
        )
    except ModelNotAvailableError as exc:
        _LOGGER.warning("Fallback generation unavailable: %s", exc, exc_info=True)
        base = current or _default_internal_config()
        return _apply_update(base, update)


async def _render_chunk(
    config_item: _MusicConfigInternal,
    *,
    chunk_seconds: float,
) -> FloatArray:
    return await asyncio.to_thread(
        lambda: ensure_audio_contract(
            assemble(_to_synth_config(config_item), chunk_seconds, normalize=False),
            sample_rate=SAMPLE_RATE,
            check_peak=False,
        )
    )


async def _prefetch_item(
    item: Streamable,
    *,
    model: ModelForGeneratingMusicConfig,
    hooks: StreamHooks | None,
    index: int,
    load_state: _ModelLoadState | None = None,
) -> _PrefetchedItem:
    task: asyncio.Task[_MusicConfigInternal] | None = None
    if isinstance(item.content, str):
        _emit_event(hooks, StreamEvent(kind="item_resolve_start", index=index, item=item))
        if load_state is None:
            task = asyncio.create_task(_generate_internal(item.content, model))
        else:
            task = asyncio.create_task(
                _generate_internal_with_load(item.content, model, load_state)
            )
    return _PrefetchedItem(index=index, item=item, task=task)


async def _resolve_target_async(
    item: StreamContent,
    *,
    current: _MusicConfigInternal | None,
    model: ModelForGeneratingMusicConfig,
    update: UpdateInput | None,
) -> _MusicConfigInternal:
    match item:
        case _MusicConfigInternal():
            target = item
        case MusicConfig():
            target = item.to_internal()
        case _MusicConfigUpdateInternal():
            base = current or _default_internal_config()
            target = (
                base
                if current is None and is_empty_update(item)
                else merge_internal_config(base, item)
            )
        case MusicConfigUpdate():
            base = current or _default_internal_config()
            internal_update = item.to_internal()
            target = (
                base
                if current is None and is_empty_update(item)
                else merge_internal_config(base, internal_update)
            )
        case str():
            target = (await model.generate(item)).to_internal()
        case _:
            raise InvalidConfigError(f"Unsupported input type: {type(item).__name__}")

    return _apply_update(target, update)


async def astream(
    items: Iterable[Streamable] | AsyncIterable[Streamable],
    *,
    chunk_seconds: float = 1.0,
    model: ModelSpec = "fast",
    config: ConfigInput | None = None,
    update: UpdateInput | None = None,
    prefetch_depth: int = 1,
    preview: bool = False,
    fallback: FallbackInput | None = None,
    fallback_model: ModelSpec = "fast",
    hooks: StreamHooks | None = None,
) -> AsyncIterable[FloatArray]:
    resolved: ModelForGeneratingMusicConfig | None = None
    fallback_resolved: ModelForGeneratingMusicConfig | None = None
    should_close = False
    fallback_should_close = False
    try:
        match chunk_seconds:
            case (float() | int()) as value if value <= 0:
                raise InvalidConfigError("chunk_seconds must be greater than 0")
            case float() | int():
                pass
            case _:
                raise InvalidConfigError("chunk_seconds must be a number")

        if prefetch_depth < 0:
            raise InvalidConfigError("prefetch_depth must be >= 0")

        current = coerce_internal_config(config) if config is not None else None
        resolved = resolve_model(model)
        fallback_resolved = resolve_model(fallback_model)
        should_close = _should_auto_close_model(model)
        fallback_should_close = _should_auto_close_model(fallback_model)
        primary_load = (
            _ModelLoadState(model=resolved, role="primary", hooks=hooks)
            if hooks is not None
            else None
        )
        fallback_load = (
            _ModelLoadState(model=fallback_resolved, role="fallback", hooks=hooks)
            if hooks is not None
            else None
        )

        _emit_event(hooks, StreamEvent(kind="stream_start"))

        async_items = _coerce_async_items(items)
        iterator = async_items.__aiter__()
        pending: deque[_PrefetchedItem] = deque()
        next_index = 0
        prefetch_target = max(1, prefetch_depth + 1)
        first_config_ready = False
        first_audio_chunk = False
        warned: dict[str, bool] = {"duration": False, "transition": False}

        async def _next_item() -> Streamable | None:
            try:
                return await iterator.__anext__()
            except StopAsyncIteration:
                return None

        async def _fill_pending() -> None:
            nonlocal next_index
            while len(pending) < prefetch_target:
                item = await _next_item()
                if item is None:
                    break
                pending.append(
                    await _prefetch_item(
                        item,
                        model=resolved,
                        hooks=hooks,
                        index=next_index,
                        load_state=primary_load,
                    )
                )
                next_index += 1

        await _fill_pending()
        while pending:
            prefetched = pending.popleft()
            await _fill_pending()

            item = prefetched.item
            index = prefetched.index
            _warn_aggressive_timing(
                duration=item.duration,
                transition_duration=item.transition_duration,
                warned=warned,
            )
            item_preview = item.preview if item.preview is not None else preview
            item_fallback = item.fallback if item.fallback is not None else fallback
            pending_task = prefetched.task

            pending_real_task: asyncio.Task[_MusicConfigInternal] | None = None
            target: _MusicConfigInternal | None = None

            if pending_task is None:
                target = await _resolve_target_async(
                    item.content,
                    current=current,
                    model=resolved,
                    update=update,
                )
                _emit_event(hooks, StreamEvent(kind="item_resolve_success", index=index, item=item))
            else:
                try:
                    target, pending_real_task = await _resolve_preview_or_primary(
                        item=item,
                        index=index,
                        pending_task=pending_task,
                        preview=item_preview,
                        fallback_model=fallback_resolved,
                        update=update,
                        fallback_load=fallback_load,
                        hooks=hooks,
                    )
                except _PendingTaskError as exc:
                    _emit_event(
                        hooks,
                        StreamEvent(
                            kind="item_resolve_error",
                            index=index,
                            item=item,
                            error=exc.error,
                            fallback=item_fallback,
                        ),
                    )
                    target = await _resolve_fallback(
                        item,
                        current=current,
                        fallback=item_fallback,
                        fallback_model=fallback_resolved,
                        update=update,
                        fallback_load=fallback_load,
                    )
                    if target is None:
                        raise
                    _emit_event(
                        hooks,
                        StreamEvent(
                            kind="fallback_used",
                            index=index,
                            item=item,
                            fallback=item_fallback,
                        ),
                    )

            if not first_config_ready:
                _emit_event(hooks, StreamEvent(kind="first_config_ready", index=index, item=item))
                first_config_ready = True

            total_chunks = _chunk_count(item.duration, chunk_seconds)
            transition_chunks = (
                0
                if current is None
                else min(
                    total_chunks,
                    _transition_steps(item.transition_duration, chunk_seconds),
                )
            )

            remaining = total_chunks
            if transition_chunks > 0:
                for config_item in _iter_transition_configs(current, target, transition_chunks):
                    current = config_item
                    if not first_audio_chunk:
                        _emit_event(hooks, StreamEvent(kind="first_audio_chunk"))
                        first_audio_chunk = True
                    yield await _render_chunk(config_item, chunk_seconds=chunk_seconds)
                remaining -= transition_chunks
            else:
                current = target

            match current:
                case _MusicConfigInternal() as current_config:
                    pass
                case None:
                    raise InvalidConfigError("Stream exhausted without a current config")
                case _:
                    raise InvalidConfigError("Stream exhausted with an invalid config")

            # ... (inside astream function) ...

            while remaining > 0:
                # 1. Create the render task but don't await it yet
                render_task = asyncio.create_task(
                    _render_chunk(current_config, chunk_seconds=chunk_seconds)
                )

                # 2. Prepare tasks to wait on
                waiting_on = {render_task}
                if pending_real_task is not None and not pending_real_task.done():
                    waiting_on.add(pending_real_task)

                # 3. Race: Wait for EITHER the render to finish OR the LLM to reply
                done, _ = await asyncio.wait(waiting_on, return_when=asyncio.FIRST_COMPLETED)

                # 4. Handle LLM completion (Priority)
                if pending_real_task in done:
                    # The LLM finished!
                    # Cancel the preview render since we have the "real" data now
                    render_task.cancel()
                    _suppress_task_exception(render_task)

                    try:
                        real_target = _apply_update(pending_real_task.result(), update)
                        _emit_event(
                            hooks,
                            StreamEvent(kind="item_resolve_success", index=index, item=item),
                        )
                    except Exception as exc:
                        _emit_event(
                            hooks,
                            StreamEvent(
                                kind="item_resolve_error",
                                index=index,
                                item=item,
                                error=exc,
                                fallback=item_fallback,
                            ),
                        )
                        # ... (existing fallback logic) ...
                        real_target = await _resolve_fallback(
                            item,
                            current=current,
                            fallback=item_fallback,
                            fallback_model=fallback_resolved,
                            update=update,
                            fallback_load=fallback_load,
                        )
                        if real_target is None:
                            raise
                        _emit_event(
                            hooks,
                            StreamEvent(
                                kind="fallback_used",
                                index=index,
                                item=item,
                                fallback=item_fallback,
                            ),
                        )

                    pending_real_task = None

                    # Calculate transition from WHERE WE ARE (current_config) to NEW TARGET
                    transition = min(
                        remaining,
                        _transition_steps(item.transition_duration, chunk_seconds),
                    )

                    if transition > 0:
                        for config_item in _iter_transition_configs(
                            current_config,
                            real_target,
                            transition,
                        ):
                            current_config = config_item
                            if not first_audio_chunk:
                                _emit_event(hooks, StreamEvent(kind="first_audio_chunk"))
                                first_audio_chunk = True
                            yield await _render_chunk(config_item, chunk_seconds=chunk_seconds)
                        remaining -= transition
                        current = current_config
                        continue

                    # No transition needed, snap to target
                    current_config = real_target
                    current = real_target
                    # Loop back to start rendering with the NEW config immediately
                    continue

                # 5. Handle Render completion (LLM didn't finish yet)
                # If we get here, the LLM wasn't the winner, so the render must be done.
                if not first_audio_chunk:
                    _emit_event(hooks, StreamEvent(kind="first_audio_chunk"))
                    first_audio_chunk = True

                yield await render_task
                remaining -= 1

            if pending_real_task is not None and not pending_real_task.done():
                pending_real_task.cancel()
                pending_real_task = None

        _emit_event(hooks, StreamEvent(kind="stream_end"))
    except Exception as exc:
        _emit_event(hooks, StreamEvent(kind="error", error=exc))
        show_console = hooks is None or hooks.on_error is None
        _log_exception("astream", exc, show_console=show_console)
        raise
    finally:
        if should_close:
            await _maybe_close_model(resolved)
        if (
            fallback_should_close
            and fallback_resolved is not None
            and fallback_resolved is not resolved
        ):
            await _maybe_close_model(fallback_resolved)



from __future__ import annotations

import asyncio
import json
import logging
import warnings
from types import SimpleNamespace
from unittest.mock import AsyncMock

import litellm
import pytest

from latentscore import MusicConfig
from latentscore.config import MusicConfigPromptPayload
from latentscore.errors import ConfigGenerateError
from latentscore.models import build_litellm_prompt
from latentscore.providers.litellm import LiteLLMAdapter, _safe_async_cleanup


def _payload_json(overrides: dict[str, object] | None = None) -> str:
    config = MusicConfig().model_dump(exclude={"schema_version", "seed"})
    if overrides:
        config.update(overrides)
    payload = {"justification": "Sound choices explained.", "config": config}
    return json.dumps(payload)


@pytest.mark.asyncio
async def test_litellm_adapter_enforces_json(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    captured: dict[str, object] = {}

    async def fake_acompletion(**kwargs: object) -> object:
        captured.update(kwargs)
        message = SimpleNamespace(content=_payload_json())
        choice = SimpleNamespace(message=message)
        return SimpleNamespace(choices=[choice])

    monkeypatch.setattr(litellm, "acompletion", fake_acompletion)

    adapter = LiteLLMAdapter(model="gemini/gemini-3-flash-preview")
    await adapter.generate("warm sunrise")

    assert captured["response_format"] is MusicConfigPromptPayload
    messages = captured["messages"]
    assert isinstance(messages, list)
    assert messages[0]["role"] == "system"
    assert messages[0]["content"] == build_litellm_prompt()


@pytest.mark.asyncio
async def test_litellm_kwargs_forwarded(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    captured: dict[str, object] = {}

    async def fake_acompletion(**kwargs: object) -> object:
        captured.update(kwargs)
        message = SimpleNamespace(content=_payload_json())
        choice = SimpleNamespace(message=message)
        return SimpleNamespace(choices=[choice])

    monkeypatch.setattr(litellm, "acompletion", fake_acompletion)

    adapter = LiteLLMAdapter(
        model="external:gemini/gemini-3-flash-preview",
        litellm_kwargs={"timeout": 42, "temperature": 0.2},
    )
    await adapter.generate("warm sunrise")

    assert captured["timeout"] == 42
    assert captured["temperature"] == 0.2


def test_litellm_adapter_reuses_loop_across_runs(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    loop_refs: list[asyncio.AbstractEventLoop] = []

    async def fake_acompletion(**kwargs: object) -> object:
        _ = kwargs
        loop_refs.append(asyncio.get_running_loop())
        message = SimpleNamespace(content=_payload_json())
        choice = SimpleNamespace(message=message)
        return SimpleNamespace(choices=[choice])

    monkeypatch.setattr(litellm, "acompletion", fake_acompletion)

    adapter = LiteLLMAdapter(model="gemini/gemini-3-flash-preview")
    asyncio.run(adapter.generate("warm sunrise"))
    asyncio.run(adapter.generate("late night neon"))

    assert len(loop_refs) == 2
    assert loop_refs[0] is loop_refs[1]


@pytest.mark.asyncio
async def test_litellm_adapter_non_json_error_includes_snippet(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    async def fake_acompletion(**kwargs: object) -> object:
        message = SimpleNamespace(content="not json output")
        choice = SimpleNamespace(message=message)
        return SimpleNamespace(choices=[choice])

    monkeypatch.setattr(litellm, "acompletion", fake_acompletion)

    adapter = LiteLLMAdapter(model="gemini/gemini-3-flash-preview")

    with pytest.raises(ConfigGenerateError) as excinfo:
        await adapter.generate("warm sunrise")

    assert "not json output" in str(excinfo.value)


@pytest.mark.asyncio
async def test_litellm_adapter_repairs_json(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    async def fake_acompletion(**kwargs: object) -> object:
        message = SimpleNamespace(content='{"justification":"ok","config":{"tempo":"slow"},}')
        choice = SimpleNamespace(message=message)
        return SimpleNamespace(choices=[choice])

    def fake_repair_json(content: str) -> str:
        _ = content
        return _payload_json({"tempo": "slow"})

    monkeypatch.setattr(litellm, "acompletion", fake_acompletion)
    monkeypatch.setattr(
        "latentscore.providers.litellm.repair_json",
        fake_repair_json,
    )

    adapter = LiteLLMAdapter(model="external:gemini/gemini-3-flash-preview")
    config = await adapter.generate("warm sunrise")

    assert config.tempo == "slow"


@pytest.mark.asyncio
async def test_litellm_adapter_accepts_raw_music_config(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    async def fake_acompletion(**kwargs: object) -> object:
        message = SimpleNamespace(content='{"tempo":"slow"}')
        choice = SimpleNamespace(message=message)
        return SimpleNamespace(choices=[choice])

    monkeypatch.setattr(litellm, "acompletion", fake_acompletion)

    adapter = LiteLLMAdapter(model="gemini/gemini-3-flash-preview")
    config = await adapter.generate("warm sunrise")

    assert config.tempo == "slow"


def test_litellm_adapter_aclose_awaits_litellm(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    aclose = AsyncMock()
    monkeypatch.setattr(litellm, "aclose", aclose, raising=False)

    adapter = LiteLLMAdapter(model="gemini/gemini-3-flash-preview")
    asyncio.run(adapter.aclose())

    aclose.assert_awaited_once()


def test_safe_async_cleanup_uses_asyncio_run_for_closed_loop() -> None:
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    loop.close()
    ran = {"value": False}

    async def cleanup() -> None:
        ran["value"] = True

    try:
        with warnings.catch_warnings(record=True) as captured:
            warnings.simplefilter("always", RuntimeWarning)
            _safe_async_cleanup(cleanup)
    finally:
        asyncio.set_event_loop(None)

    assert ran["value"] is True
    assert not [warn for warn in captured if issubclass(warn.category, RuntimeWarning)]


def test_safe_async_cleanup_logs_failures(caplog: pytest.LogCaptureFixture) -> None:
    async def cleanup() -> None:
        raise RuntimeError("boom")

    caplog.set_level(logging.WARNING)
    _safe_async_cleanup(cleanup)

    assert any("boom" in record.getMessage() for record in caplog.records)


def test_cleanup_message_not_emitted_at_info(
    caplog: pytest.LogCaptureFixture,
) -> None:
    async def cleanup() -> None:
        return None

    caplog.set_level(logging.INFO, logger="latentscore.providers.litellm")
    _safe_async_cleanup(cleanup)

    assert all(
        "LiteLLM cleanup running outside an event loop." not in record.getMessage()
        for record in caplog.records
    )



from __future__ import annotations

import os
import sys
import threading
import time
import traceback
from contextlib import contextmanager
from typing import IO, TYPE_CHECKING, Any, Iterator

if TYPE_CHECKING:
    from rich.console import Console as RichConsole
    from rich.progress import Progress as RichProgress
    from rich.progress import TaskID as RichTaskID
    from rich.status import Status as RichStatus
else:
    RichConsole = Any
    RichProgress = Any
    RichTaskID = Any
    RichStatus = Any

try:
    from rich.console import Console
    from rich.panel import Panel
    from rich.progress import (
        BarColumn,
        Progress,
        SpinnerColumn,
        TextColumn,
        TimeElapsedColumn,
        TimeRemainingColumn,
    )
    from rich.status import Status
    from rich.text import Text
    from rich.traceback import Traceback
except ImportError:  # pragma: no cover - optional dependency
    Console = None
    Panel = None
    Progress = None
    Status = None
    Text = None
    Traceback = None
    BarColumn = None
    TextColumn = None
    SpinnerColumn = None
    TimeElapsedColumn = None
    TimeRemainingColumn = None


class _AsciiSpinner:
    def __init__(
        self,
        message: str,
        *,
        interval: float = 0.1,
        frames: str = "|/-\\",
        stream: IO[str] | None = None,
        enabled: bool | None = None,
    ) -> None:
        self._message = message
        self._interval = interval
        self._frames = frames
        self._stream = stream or sys.stderr
        self._enabled = self._stream.isatty() if enabled is None else enabled
        self._stop = threading.Event()
        self._thread: threading.Thread | None = None
        self._lock = threading.Lock()
        self._last_len = 0

    def start(self) -> None:
        if not self._enabled:
            return
        if self._thread and self._thread.is_alive():
            return
        self._stop.clear()
        self._thread = threading.Thread(target=self._run, daemon=True)
        self._thread.start()

    def update(self, message: str) -> None:
        with self._lock:
            self._message = message

    def stop(self) -> None:
        if not self._enabled:
            return
        self._stop.set()
        if self._thread:
            self._thread.join(timeout=0.2)
        self._clear_line()

    def _run(self) -> None:
        index = 0
        while not self._stop.is_set():
            frame = self._frames[index % len(self._frames)]
            message = self._current_message()
            text = f"{message} {frame}"
            self._render(text)
            index += 1
            time.sleep(self._interval)
        self._clear_line()

    def _current_message(self) -> str:
        with self._lock:
            return self._message

    def _render(self, text: str) -> None:
        self._last_len = max(self._last_len, len(text))
        self._stream.write(f"\r{text}")
        self._stream.flush()

    def _clear_line(self) -> None:
        if self._last_len <= 0:
            return
        self._stream.write("\r" + (" " * self._last_len) + "\r")
        self._stream.flush()


class _RichSpinner:
    def __init__(
        self,
        message: str,
        *,
        spinner: str | None = None,
        stream: IO[str] | None = None,
    ) -> None:
        self._message = message
        self._spinner = spinner or "dots"
        self._stream = stream or sys.stderr
        self._console: RichConsole | None = Console(file=self._stream) if Console else None
        self._status: RichStatus | None = None

    def start(self) -> None:
        if self._console is None or Status is None:
            return
        if self._status is not None:
            return
        self._status = self._console.status(self._message, spinner=self._spinner)
        self._status.start()

    def update(self, message: str) -> None:
        self._message = message
        if self._status is not None:
            self._status.update(message)

    def stop(self) -> None:
        if self._status is None:
            return
        self._status.stop()
        self._status = None


class _RichElapsedSpinner:
    def __init__(
        self,
        message: str,
        *,
        spinner: str | None = None,
        stream: IO[str] | None = None,
    ) -> None:
        self._message = message
        self._spinner = spinner or "dots"
        self._stream = stream or sys.stderr
        self._console: RichConsole | None = Console(file=self._stream) if Console else None
        self._progress: RichProgress | None = None
        self._task_id: RichTaskID | None = None

    def start(self) -> None:
        if (
            self._console is None
            or Progress is None
            or SpinnerColumn is None
            or TextColumn is None
            or TimeElapsedColumn is None
        ):
            return
        if self._progress is not None:
            return
        self._progress = Progress(
            SpinnerColumn(spinner_name=self._spinner),
            TextColumn("[progress.description]{task.description}"),
            TimeElapsedColumn(),
            console=self._console,
            transient=True,
        )
        self._progress.start()
        self._task_id = self._progress.add_task(self._message, total=None)

    def update(self, message: str) -> None:
        self._message = message
        if self._progress is not None and self._task_id is not None:
            self._progress.update(self._task_id, description=message)

    def stop(self) -> None:
        if self._progress is None:
            return
        self._progress.stop()
        self._progress = None
        self._task_id = None


class Spinner:
    """Spinner helper with Rich status fallback."""

    def __init__(
        self,
        message: str,
        *,
        interval: float = 0.1,
        frames: str | None = None,
        spinner: str | None = None,
        stream: IO[str] | None = None,
        enabled: bool | None = None,
        show_elapsed: bool = False,
    ) -> None:
        self._message = message
        self._interval = interval
        self._stream = stream or sys.stderr
        self._enabled = self._stream.isatty() if enabled is None else enabled
        self._backend: _AsciiSpinner | _RichSpinner | _RichElapsedSpinner | None = None
        if not self._enabled:
            return
        if Console:
            if show_elapsed and Progress and SpinnerColumn and TextColumn and TimeElapsedColumn:
                self._backend = _RichElapsedSpinner(
                    self._message,
                    spinner=spinner,
                    stream=self._stream,
                )
            elif Status:
                self._backend = _RichSpinner(
                    self._message,
                    spinner=spinner,
                    stream=self._stream,
                )
            else:
                self._backend = _AsciiSpinner(
                    self._message,
                    interval=self._interval,
                    frames=frames or "|/-\\",
                    stream=self._stream,
                    enabled=self._enabled,
                )
        else:
            self._backend = _AsciiSpinner(
                self._message,
                interval=self._interval,
                frames=frames or "|/-\\",
                stream=self._stream,
                enabled=self._enabled,
            )

    def start(self) -> None:
        if self._backend is None:
            return
        self._backend.start()

    def update(self, message: str) -> None:
        self._message = message
        if self._backend is None:
            return
        self._backend.update(message)

    def stop(self) -> None:
        if self._backend is None:
            return
        self._backend.stop()

    def __enter__(self) -> "Spinner":
        self.start()
        return self

    def __exit__(self, exc_type: object, exc: object, tb: object) -> None:
        self.stop()


@contextmanager
def spinner(
    message: str,
    *,
    interval: float = 0.1,
    frames: str | None = None,
    spinner: str | None = None,
    stream: IO[str] | None = None,
    enabled: bool | None = None,
    show_elapsed: bool = False,
) -> Iterator[Spinner]:
    handle = Spinner(
        message,
        interval=interval,
        frames=frames,
        spinner=spinner,
        stream=stream,
        enabled=enabled,
        show_elapsed=show_elapsed,
    )
    handle.start()
    try:
        yield handle
    finally:
        handle.stop()


class _AsciiProgressBar:
    def __init__(
        self,
        message: str,
        *,
        total: float,
        stream: IO[str] | None = None,
        enabled: bool | None = None,
        width: int = 24,
    ) -> None:
        self._message = message
        self._total = total
        self._stream = stream or sys.stderr
        self._enabled = self._stream.isatty() if enabled is None else enabled
        self._width = width
        self._last_len = 0

    def start(self) -> None:
        if not self._enabled:
            return
        self.update(0.0)

    def update(self, completed: float) -> None:
        if not self._enabled:
            return
        if self._total <= 0:
            fraction = 1.0
        else:
            fraction = max(0.0, min(completed / self._total, 1.0))
        filled = int(round(self._width * fraction))
        bar = "=" * filled + "-" * (self._width - filled)
        percent = int(round(100 * fraction))
        text = f"{self._message} [{bar}] {percent:>3d}%"
        self._render(text)

    def stop(self) -> None:
        if not self._enabled:
            return
        self._clear_line()

    def _render(self, text: str) -> None:
        self._last_len = max(self._last_len, len(text))
        self._stream.write(f"\r{text}")
        self._stream.flush()

    def _clear_line(self) -> None:
        if self._last_len <= 0:
            return
        self._stream.write("\r" + (" " * self._last_len) + "\r")
        self._stream.flush()


class _RichProgressBar:
    def __init__(
        self,
        message: str,
        *,
        total: float,
        stream: IO[str] | None = None,
    ) -> None:
        self._message = message
        self._total = total
        self._stream = stream or sys.stderr
        self._console: RichConsole | None = Console(file=self._stream) if Console else None
        self._progress: RichProgress | None = None
        self._task_id: RichTaskID | None = None

    def start(self) -> None:
        if (
            self._console is None
            or Progress is None
            or TextColumn is None
            or BarColumn is None
            or TimeElapsedColumn is None
            or TimeRemainingColumn is None
        ):
            return
        if self._progress is not None:
            return
        self._progress = Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TimeElapsedColumn(),
            TimeRemainingColumn(),
            console=self._console,
            transient=True,
        )
        self._progress.start()
        self._task_id = self._progress.add_task(self._message, total=self._total)

    def update(self, completed: float) -> None:
        if self._progress is None or self._task_id is None:
            return
        self._progress.update(self._task_id, completed=min(completed, self._total))

    def stop(self) -> None:
        if self._progress is None:
            return
        self._progress.stop()
        self._progress = None
        self._task_id = None


class ProgressBar:
    """Progress bar helper with Rich fallback."""

    def __init__(
        self,
        message: str,
        *,
        total: float,
        stream: IO[str] | None = None,
        enabled: bool | None = None,
    ) -> None:
        self._message = message
        self._total = total
        self._stream = stream or sys.stderr
        self._enabled = self._stream.isatty() if enabled is None else enabled
        self._backend: _AsciiProgressBar | _RichProgressBar | None = None
        if not self._enabled:
            return
        if Console and Progress:
            self._backend = _RichProgressBar(
                self._message,
                total=self._total,
                stream=self._stream,
            )
        else:
            self._backend = _AsciiProgressBar(
                self._message,
                total=self._total,
                stream=self._stream,
                enabled=self._enabled,
            )

    def start(self) -> None:
        if self._backend is None:
            return
        self._backend.start()

    def update(self, completed: float) -> None:
        if self._backend is None:
            return
        self._backend.update(completed)

    def stop(self) -> None:
        if self._backend is None:
            return
        self._backend.stop()


def render_error(
    context: str,
    exc: BaseException,
    *,
    stream: IO[str] | None = None,
) -> None:
    target = stream or sys.stderr
    debug = os.environ.get("LATENTSCORE_DEBUG")
    log_path = None
    try:
        from .logging_utils import get_log_path

        log_path = get_log_path()
    except Exception:
        log_path = None
    if Console and Panel and Text and Traceback and target.isatty():
        console = Console(file=target)
        headline = Text(f"{type(exc).__name__}", style="bold red")
        summary = Text(str(exc))
        log_line = ""
        if log_path is not None:
            log_line = f"\nLogs: {log_path}"
        body = Text.assemble(
            ("LatentScore error while ", "bold"),
            (context, "bold"),
            (":\n\n", "bold"),
            headline,
            (": ", "bold"),
            summary,
            (log_line, "dim"),
            ("\n\nSet LATENTSCORE_DEBUG=1 for console trace.", "dim"),
        )
        console.print(Panel(body, title="Error", border_style="red"))
        if debug:
            console.print(Traceback.from_exception(type(exc), exc, exc.__traceback__))
    else:
        suffix = f" (logs: {log_path})" if log_path is not None else ""
        target.write(f"{context} failed: {type(exc).__name__}: {exc}{suffix}\n")
        if debug:
            traceback.print_exception(type(exc), exc, exc.__traceback__, file=target)



from __future__ import annotations

import logging
import os
from collections.abc import AsyncIterable, AsyncIterator, Callable, Iterable, Iterator, Sequence
from pathlib import Path
from typing import Any, Literal

import numpy as np
from numpy.typing import DTypeLike, NDArray
from pydantic import BaseModel, ConfigDict, model_validator

from .audio import SAMPLE_RATE, FloatArray, ensure_audio_contract, write_wav
from .config import (
    ConfigInput,
    MusicConfig,
    MusicConfigUpdate,
    SynthConfig,
    UpdateInput,
    coerce_internal_config,
    coerce_internal_update,
    merge_internal_config,
)
from .errors import InvalidConfigError, ModelNotAvailableError, PlaybackError
from .indicators import RichIndicator
from .main import FallbackInput, RenderHooks, Streamable, StreamHooks
from .main import (
    astream as astream_raw,
)
from .main import (
    render as render_raw,
)
from .main import (
    stream as stream_raw,
)
from .models import EXTERNAL_PREFIX, ModelSpec
from .spinner import render_error
from .synth import assemble

TrackContent = str | MusicConfig | MusicConfigUpdate

_LOGGER = logging.getLogger("latentscore.dx")


class Audio(BaseModel):
    samples: FloatArray
    sample_rate: int = SAMPLE_RATE

    model_config = ConfigDict(
        frozen=True,
        extra="forbid",
        arbitrary_types_allowed=True,
    )

    @model_validator(mode="after")
    def _normalize(self) -> "Audio":
        normalized = ensure_audio_contract(self.samples, sample_rate=self.sample_rate)
        object.__setattr__(self, "samples", normalized)
        return self

    def to_numpy(self) -> FloatArray:
        return self.samples

    def __array__(self, dtype: DTypeLike | None = None) -> NDArray[np.generic]:
        return np.asarray(self.samples, dtype=dtype)

    def save(self, path: str | Path) -> Path:
        return write_wav(path, self.samples, sample_rate=self.sample_rate)

    def play(self) -> None:
        from .playback import play_audio

        try:
            play_audio(self.samples, sample_rate=self.sample_rate)
        except PlaybackError as exc:
            render_error("playback", exc)
            raise


class AudioStream:
    def __init__(
        self,
        sync_factory: Callable[[], Iterable[FloatArray]],
        async_factory: Callable[[], AsyncIterable[FloatArray]],
        *,
        sample_rate: int = SAMPLE_RATE,
    ) -> None:
        self._sync_factory = sync_factory
        self._async_factory = async_factory
        self.sample_rate = sample_rate

    def __iter__(self) -> Iterator[FloatArray]:
        return iter(self._sync_factory())

    def __aiter__(self) -> AsyncIterator[FloatArray]:
        return self._async_factory().__aiter__()

    def collect(self) -> Audio:
        chunks = [
            ensure_audio_contract(chunk, sample_rate=self.sample_rate)
            for chunk in self._sync_factory()
        ]
        if not chunks:
            return Audio(samples=np.array([], dtype=np.float32), sample_rate=self.sample_rate)
        return Audio(samples=np.concatenate(chunks), sample_rate=self.sample_rate)

    def save(self, path: str | Path) -> Path:
        return write_wav(path, self._sync_factory(), sample_rate=self.sample_rate)

    def play(self) -> None:
        from .playback import play_stream

        try:
            play_stream(self._sync_factory(), sample_rate=self.sample_rate)
        except PlaybackError:
            _LOGGER.info(
                "Streaming playback unavailable; falling back to buffered play.", exc_info=True
            )
            self.collect().play()


class Track(BaseModel):
    content: TrackContent
    duration: float = 60.0
    transition: float = 5.0
    name: str | None = None

    model_config = ConfigDict(
        frozen=True,
        extra="forbid",
        arbitrary_types_allowed=True,
    )

    def to_streamable(self) -> Streamable:
        return Streamable(
            content=self.content,
            duration=self.duration,
            transition_duration=self.transition,
        )


StreamItems = Track | TrackContent
StreamItemsInput = StreamItems | Sequence[StreamItems]


class Playlist(BaseModel):
    tracks: tuple[Track, ...]

    model_config = ConfigDict(
        frozen=True,
        extra="forbid",
        arbitrary_types_allowed=True,
    )

    def stream(
        self,
        *,
        chunk_seconds: float = 1.0,
        model: ModelSpec = "fast",
        config: ConfigInput | None = None,
        update: UpdateInput | None = None,
        prefetch_depth: int = 1,
        preview: bool = False,
        fallback: FallbackInput | None = None,
        fallback_model: ModelSpec = "fast",
        hooks: StreamHooks | None = None,
        queue_maxsize: int = 0,
    ) -> AudioStream:
        return _stream_from_tracks(
            list(self.tracks),
            chunk_seconds=chunk_seconds,
            model=model,
            config=config,
            update=update,
            prefetch_depth=prefetch_depth,
            preview=preview,
            fallback=fallback,
            fallback_model=fallback_model,
            hooks=hooks,
            queue_maxsize=queue_maxsize,
        )

    def render(self, **kwargs: Any) -> Audio:
        return self.stream(**kwargs).collect()

    def play(self, **kwargs: Any) -> None:
        self.stream(**kwargs).play()


def render(
    vibe_or_config: TrackContent,
    *,
    duration: float = 8.0,
    model: ModelSpec = "fast",
    config: ConfigInput | None = None,
    update: UpdateInput | None = None,
    sample_rate: int = SAMPLE_RATE,
    hooks: RenderHooks | None = None,
) -> Audio:
    if sample_rate != SAMPLE_RATE:
        raise InvalidConfigError(f"sample_rate must be {SAMPLE_RATE}")

    resolved_hooks = hooks
    if hooks is None:
        resolved_hooks = RichIndicator().render_hooks()

    if isinstance(vibe_or_config, str):
        resolved_model = _coerce_model(model)
        audio = render_raw(
            vibe_or_config,
            duration=duration,
            model=resolved_model,
            config=config,
            update=update,
            hooks=resolved_hooks,
        )
        return Audio(samples=audio, sample_rate=sample_rate)

    base = coerce_internal_config(config) if config is not None else MusicConfig().to_internal()
    match vibe_or_config:
        case MusicConfig():
            target = vibe_or_config.to_internal()
        case MusicConfigUpdate():
            internal_update = coerce_internal_update(vibe_or_config)
            target = merge_internal_config(base, internal_update)
        case _:
            raise InvalidConfigError(
                f"Unsupported render input type: {type(vibe_or_config).__name__}"
            )

    if update is not None:
        target = merge_internal_config(target, coerce_internal_update(update))

    _emit_render_hooks(resolved_hooks, kind="start")
    try:
        synth_config = SynthConfig.model_validate(target.model_dump())
        _emit_render_hooks(resolved_hooks, kind="synth_start")
        audio = assemble(synth_config, duration=duration)
        normalized = ensure_audio_contract(audio, sample_rate=sample_rate)
        _emit_render_hooks(resolved_hooks, kind="synth_end")
        _emit_render_hooks(resolved_hooks, kind="end")
        return Audio(samples=normalized, sample_rate=sample_rate)
    except Exception as exc:
        _emit_render_hooks(resolved_hooks, kind="error", error=exc)
        debug = bool(os.environ.get("LATENTSCORE_DEBUG"))
        _LOGGER.warning("render failed: %s", exc, exc_info=debug)
        if hooks is not None and hooks.on_error is None:
            render_error("render", exc)
        raise


def stream(
    *items: StreamItemsInput,
    duration: float = 60.0,
    transition: float = 5.0,
    chunk_seconds: float = 1.0,
    model: ModelSpec = "fast",
    config: ConfigInput | None = None,
    update: UpdateInput | None = None,
    prefetch_depth: int = 1,
    preview: bool = False,
    fallback: FallbackInput | None = None,
    fallback_model: ModelSpec = "fast",
    hooks: StreamHooks | None = None,
    queue_maxsize: int = 0,
) -> AudioStream:
    items_list = _normalize_items(items)
    tracks = _coerce_tracks(items_list, duration=duration, transition=transition)
    return _stream_from_tracks(
        tracks,
        chunk_seconds=chunk_seconds,
        model=model,
        config=config,
        update=update,
        prefetch_depth=prefetch_depth,
        preview=preview,
        fallback=fallback,
        fallback_model=fallback_model,
        hooks=hooks,
        queue_maxsize=queue_maxsize,
    )


def _normalize_items(items: tuple[StreamItemsInput, ...]) -> list[StreamItems]:
    if (
        len(items) == 1
        and isinstance(items[0], Sequence)
        and not isinstance(
            items[0],
            (str, bytes),
        )
    ):
        sequence = items[0]
        items_list = list(sequence)
    else:
        items_list: list[StreamItems] = []
        for item in items:
            if isinstance(item, Sequence) and not isinstance(item, (str, bytes)):
                raise InvalidConfigError("stream expects items or a single sequence")
            items_list.append(item)

    if not items_list:
        raise InvalidConfigError("stream expects at least one item")

    return items_list


def _coerce_tracks(
    items: Sequence[StreamItems],
    *,
    duration: float,
    transition: float,
) -> list[Track]:
    count = len(items)
    per_track = duration / count
    tracks: list[Track] = []
    for item in items:
        if isinstance(item, Track):
            tracks.append(item)
        else:
            tracks.append(Track(content=item, duration=per_track, transition=transition))
    return tracks


def _streamables_for_tracks(tracks: Sequence[Track]) -> Iterable[Streamable]:
    for track in tracks:
        yield track.to_streamable()


def _emit_render_hooks(
    hooks: RenderHooks | None,
    *,
    kind: Literal["start", "synth_start", "synth_end", "end", "error"],
    error: Exception | None = None,
) -> None:
    if hooks is None:
        return
    try:
        match kind:
            case "start":
                if hooks.on_start is not None:
                    hooks.on_start()
            case "synth_start":
                if hooks.on_synth_start is not None:
                    hooks.on_synth_start()
            case "synth_end":
                if hooks.on_synth_end is not None:
                    hooks.on_synth_end()
            case "end":
                if hooks.on_end is not None:
                    hooks.on_end()
            case "error":
                if hooks.on_error is not None and error is not None:
                    hooks.on_error(error)
            case _:
                raise InvalidConfigError(f"Unknown render hook kind: {kind}")
    except Exception as exc:
        debug = bool(os.environ.get("LATENTSCORE_DEBUG"))
        _LOGGER.warning("Render hook failed: %s", exc, exc_info=debug)


def _stream_from_tracks(
    tracks: list[Track],
    *,
    chunk_seconds: float,
    model: ModelSpec,
    config: ConfigInput | None,
    update: UpdateInput | None,
    prefetch_depth: int,
    preview: bool,
    fallback: FallbackInput | None,
    fallback_model: ModelSpec,
    hooks: StreamHooks | None,
    queue_maxsize: int,
) -> AudioStream:
    resolved_model = _coerce_model(model)
    resolved_fallback = _coerce_model(fallback_model)
    resolved_hooks = hooks
    indicator = None
    if hooks is None:
        indicator = RichIndicator()
        resolved_hooks = indicator.stream_hooks()

    def sync_factory() -> Iterable[FloatArray]:
        _ = indicator
        return stream_raw(
            _streamables_for_tracks(tracks),
            chunk_seconds=chunk_seconds,
            model=resolved_model,
            config=config,
            update=update,
            prefetch_depth=prefetch_depth,
            preview=preview,
            fallback=fallback,
            fallback_model=resolved_fallback,
            hooks=resolved_hooks,
            queue_maxsize=queue_maxsize,
        )

    def async_factory() -> AsyncIterable[FloatArray]:
        _ = indicator
        return astream_raw(
            _streamables_for_tracks(tracks),
            chunk_seconds=chunk_seconds,
            model=resolved_model,
            config=config,
            update=update,
            prefetch_depth=prefetch_depth,
            preview=preview,
            fallback=fallback,
            fallback_model=resolved_fallback,
            hooks=resolved_hooks,
        )

    return AudioStream(sync_factory, async_factory, sample_rate=SAMPLE_RATE)


def _coerce_model(model: ModelSpec) -> ModelSpec:
    if isinstance(model, str) and model.startswith(EXTERNAL_PREFIX):
        name = model[len(EXTERNAL_PREFIX) :].strip()
        if not name:
            raise InvalidConfigError("external model string must be 'external:<model-name>'")
        try:
            from .providers.litellm import LiteLLMAdapter
        except ImportError as exc:
            _LOGGER.warning("LiteLLM not installed: %s", exc, exc_info=True)
            raise ModelNotAvailableError(
                "LiteLLM is not installed; install latentscore[litellm]"
            ) from exc
        return LiteLLMAdapter(model=name)

    return model



from __future__ import annotations

import argparse
import functools
import json
import logging
import os
import re
from pathlib import Path
from typing import Iterable, NamedTuple

from pydantic import ValidationError

import latentscore as ls
from latentscore.config import MusicConfig, MusicConfigPromptPayload
from latentscore.errors import ModelNotAvailableError
from latentscore.models import MODEL_CHOICES, ExpressiveMlxModel, ModelChoice
from latentscore.playback import play_stream
from latentscore.prompt_examples import FEW_SHOT_EXAMPLES as TRACK_EXAMPLES_TEXT

ROOT = Path(__file__).resolve().parents[1]
OUTPUT_DIR = ROOT / ".examples"
EXTERNAL_MODEL = "external:gemini/gemini-3-flash-preview"
EXTERNAL_API_ENV = "GEMINI_API_KEY"
# EXTERNAL_MODEL = "external:anthropic/claude-opus-4-5-20251101"
# EXTERNAL_API_ENV = "ANTHROPIC_API_KEY"
# EXTERNAL_MODEL = "external:openrouter/mistralai/voxtral-small-24b-2507"
# EXTERNAL_API_ENV = "OPENROUTER_API_KEY"
TRACK_DURATION = 12.0
DEFAULT_VIBE = "warm sunrise over water"
DEFAULT_MODEL: ModelChoice = "fast"
_LOGGER = logging.getLogger("latentscore.demo")

_TRACK_EXAMPLE_PATTERN = re.compile(
    r"\*\*Example\s+(?P<num>\d+)\*\*\s*Input:\s*\"(?P<input>.*?)\"\s*Output:\s*\n\n```json\n(?P<json>.*?)\n```",
    re.DOTALL,
)


class TrackExample(NamedTuple):
    example_index: int
    key: str
    label: str
    input_text: str
    config: MusicConfig


def _slugify(text: str) -> str:
    normalized = re.sub(r"[^a-z0-9]+", "-", text.lower())
    return normalized.strip("-")


def _parse_track_config(payload: str) -> MusicConfig:
    try:
        wrapper = MusicConfigPromptPayload.model_validate_json(payload)
        return wrapper.config.to_config()
    except ValidationError:
        try:
            return MusicConfig.model_validate_json(payload)
        except ValidationError:
            data = json.loads(payload)
            match data:
                case {"config": config} if isinstance(config, dict):
                    return MusicConfig.model_validate(config)
                case _:
                    raise ValueError("Track example missing config payload.") from None


@functools.lru_cache(maxsize=1)
def _track_examples() -> tuple[TrackExample, ...]:
    examples: list[TrackExample] = []
    for match in _TRACK_EXAMPLE_PATTERN.finditer(TRACK_EXAMPLES_TEXT):
        example_index = int(match.group("num"))
        input_text = match.group("input").strip()
        label = input_text.split(" - ", 1)[0].strip()
        key = _slugify(label) or f"example-{example_index}"
        payload = match.group("json").strip()
        config = _parse_track_config(payload)
        examples.append(
            TrackExample(
                example_index=example_index,
                key=key,
                label=label,
                input_text=input_text,
                config=config,
            )
        )
    return tuple(examples)


def _track_index() -> dict[str, TrackExample]:
    index: dict[str, TrackExample] = {}
    for example in _track_examples():
        keys = {
            example.key,
            example.label.lower(),
            example.input_text.lower(),
            str(example.example_index),
            f"example-{example.example_index}",
        }
        for key in keys:
            if key and key not in index:
                index[key] = example
    return index


def _resolve_track_example(track: str) -> TrackExample | None:
    query = track.strip()
    if not query:
        return None
    index = _track_index()
    normalized = query.lower()
    for key in (normalized, _slugify(query)):
        if key in index:
            return index[key]
    return None


def _print_track_examples() -> None:
    examples = _track_examples()
    if not examples:
        print("No track examples available.")
        return
    print("Available track examples:")
    for example in examples:
        print(f"- example-{example.example_index}: {example.label} ({example.key})")


def _live_playlist_generator() -> Iterable[ls.Streamable]:
    DURATION, TRANSITION_DURATION = 6.0, 1.5
    yield ls.Streamable(
        content="misty harbor at dawn",
        duration=DURATION,
        transition_duration=TRANSITION_DURATION,
    )
    yield ls.Streamable(
        content="neon skyline rain",
        duration=DURATION,
        transition_duration=TRANSITION_DURATION,
    )
    yield ls.Streamable(
        content="quiet orchard dusk",
        duration=DURATION,
        transition_duration=TRANSITION_DURATION,
    )
    yield ls.Streamable(
        content="starlit highway",
        duration=DURATION,
        transition_duration=TRANSITION_DURATION,
    )


def _demo_live_from_generator(
    model: ModelChoice,
    *,
    preview: bool = False,
    fallback_model: ModelChoice = "fast",
) -> None:
    live_chunks = ls.stream_raw(
        _live_playlist_generator(),
        chunk_seconds=1.0,
        model=model,
        preview=preview,
        fallback_model=fallback_model,
    )
    play_stream(live_chunks, sample_rate=ls.SAMPLE_RATE)


def _preview_enabled_for_stream(
    model: ModelChoice,
    *,
    speculative: bool,
) -> bool:
    if not speculative:
        return False
    if model == "fast":
        return False
    return True


def _resolve_live_model(model: ModelChoice) -> ModelChoice:
    if model not in ("expressive", "local"):
        return model
    try:
        ExpressiveMlxModel.check_dependencies()
    except ModelNotAvailableError as exc:
        _LOGGER.warning("Expressive model unavailable: %s; falling back to fast.", exc)
        return "fast"
    return model


def _demo_render_vibe(vibe: str, *, model: ModelChoice, save: bool) -> None:
    audio = ls.render(
        vibe,
        duration=8.0,
        model=model,
    )
    audio.play()
    if save:
        audio.save(OUTPUT_DIR / "demo_render.wav")


def _demo_external_with_key(model: str, api_key: str, *, save: bool) -> None:
    stream = ls.stream(
        "vinyl crackle midnight jazz",
        chunk_seconds=2,
        duration=30.0,
        model=ls.ExternalModelSpec(
            model=model, api_key=api_key, litellm_kwargs={"temperature": 0.0}
        ),
        # model="fast",
        # preview=True,
    )
    stream.play()
    if save:
        stream.save(OUTPUT_DIR / "demo_external_key.wav")


def _demo_external_with_env(model: str, api_env: str, *, save: bool) -> bool:
    api_key = os.environ.get(api_env, "").strip()
    if not api_key:
        return False
    os.environ[api_env] = api_key
    audio = ls.render(
        "vinyl crackle midnight jazz",
        duration=30.0,
        model=ls.ExternalModelSpec(model=model, api_key=None),
    )
    audio.play()
    if save:
        audio.save(OUTPUT_DIR / "demo_external_env.wav")
    return True


def _demo_track_example(example: TrackExample, *, save: bool) -> None:
    audio = ls.render(
        example.input_text,
        duration=TRACK_DURATION,
        config=example.config,
    )
    audio.play()
    if save:
        audio.save(OUTPUT_DIR / f"demo_track_{example.key}.wav")


def _parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog="latentscore.demo")
    mode_group = parser.add_mutually_exclusive_group()
    mode_group.add_argument("--live", action="store_true", help="Run the live generator stream.")
    mode_group.add_argument(
        "--external",
        action="store_true",
        help="Run external LLM demos.",
    )
    mode_group.add_argument(
        "--track",
        type=str,
        default=None,
        help="Render a named few-shot track using LiteLLM (see --list-tracks).",
    )
    parser.add_argument(
        "--external-model",
        type=str,
        default=EXTERNAL_MODEL,
        help=f"External model name (default: {EXTERNAL_MODEL}).",
    )
    parser.add_argument(
        "--model",
        type=str,
        choices=MODEL_CHOICES,
        default=DEFAULT_MODEL,
        help=f"Model for non-external demos (default: {DEFAULT_MODEL}).",
    )
    parser.add_argument(
        "--speculative",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="Enable speculative streaming warmup (only applies with --live).",
    )
    parser.add_argument(
        "--vibe",
        type=str,
        default=DEFAULT_VIBE,
        help=f"Prompt to render when no mode is selected (default: {DEFAULT_VIBE!r}).",
    )
    parser.add_argument(
        "--api-key",
        type=str,
        default=None,
        help="API key for the explicit external LLM demo.",
    )
    parser.add_argument(
        "--api-env",
        type=str,
        default=EXTERNAL_API_ENV,
        help=f"Environment variable name for API key (default: {EXTERNAL_API_ENV}).",
    )
    parser.add_argument(
        "--list-tracks",
        action="store_true",
        help="List available track names for --track.",
    )
    parser.add_argument(
        "--save",
        action="store_true",
        help=f"Save demo audio into {OUTPUT_DIR}.",
    )
    return parser.parse_args(argv)


def main(argv: list[str] | None = None) -> None:
    args = _parse_args(argv)
    if args.list_tracks:
        _print_track_examples()
        return
    if args.save:
        OUTPUT_DIR.mkdir(parents=True, exist_ok=True)

    if args.live:
        resolved_model = _resolve_live_model(args.model)
        preview = _preview_enabled_for_stream(resolved_model, speculative=args.speculative)
        if preview:
            _LOGGER.warning(
                "Speculative streaming enabled (fast model warmup while expressive loads). "
                "Use --no-speculative to wait for the expressive model."
            )
        _demo_live_from_generator(
            resolved_model,
            preview=preview,
            fallback_model="fast",
        )
        return

    if args.external:
        ran = False
        api_key = (
            args.api_key.strip() if args.api_key else os.environ.get(args.api_env, "").strip() or ""
        )
        if api_key:
            _demo_external_with_key(args.external_model, api_key, save=args.save)
            ran = True

        if not api_key:
            if _demo_external_with_env(args.external_model, args.api_env, save=args.save):
                ran = True

        if not ran:
            print(
                f"No external demos executed. Provide --api-key your_key_here or set {args.api_env} to run them."
            )
        return

    if args.track:
        example = _resolve_track_example(args.track)
        if example is None:
            print(f"Unknown track: {args.track}")
            _print_track_examples()
            return
        _demo_track_example(example, save=args.save)
        return

    _demo_render_vibe(args.vibe, model=args.model, save=args.save)


if __name__ == "__main__":
    main()



from __future__ import annotations

import logging
import os
import sys
import traceback
from datetime import datetime
from pathlib import Path

_LOGGER = logging.getLogger("latentscore.logging")
_LOG_DIR_ENV = "LATENTSCORE_LOG_DIR"
_MODEL_DIR_ENV = "LATENTSCORE_MODEL_DIR"
_LOG_FILE = "latentscore.log"
_logging_configured = False
_CONSOLE_FORMAT = "%(level_prefix)s %(name)s: %(message)s"
_FILE_FORMAT = "%(asctime)s [%(levelname)s] %(name)s: %(message)s"
_FILE_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
_LEVEL_PREFIXES = {
    logging.DEBUG: "🐛",
    logging.INFO: "ℹ️",
    logging.WARNING: "⚠️",
    logging.ERROR: "❌",
    logging.CRITICAL: "💥",
}


def _level_prefix(levelno: int) -> str:
    return _LEVEL_PREFIXES.get(levelno, "")


class _ConsoleEmojiFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        record.level_prefix = _level_prefix(record.levelno)
        return super().format(record)


def _default_model_dir() -> Path:
    configured = os.environ.get(_MODEL_DIR_ENV)
    if configured:
        return Path(configured).expanduser()
    return Path.home() / ".cache" / "latentscore" / "models"


def get_log_dir() -> Path:
    configured = os.environ.get(_LOG_DIR_ENV)
    if configured:
        return Path(configured).expanduser()
    return _default_model_dir().parent / "logs"


def get_log_path() -> Path:
    return get_log_dir() / _LOG_FILE


def configure_logging(*, force: bool = False) -> None:
    global _logging_configured
    if _logging_configured and not force:
        return

    logger = logging.getLogger("latentscore")
    logger.setLevel(logging.DEBUG)

    if force:
        for handler in list(logger.handlers):
            logger.removeHandler(handler)

    root_has_handlers = bool(logging.getLogger().handlers)
    if force or not root_has_handlers:
        console_handler = logging.StreamHandler(stream=sys.__stderr__)
        console_level = logging.DEBUG if os.environ.get("LATENTSCORE_DEBUG") else logging.INFO
        console_handler.setLevel(console_level)
        console_handler.setFormatter(_ConsoleEmojiFormatter(_CONSOLE_FORMAT))
        logger.addHandler(console_handler)

    try:
        log_dir = get_log_dir()
        log_dir.mkdir(parents=True, exist_ok=True)
        file_handler = logging.FileHandler(get_log_path(), encoding="utf-8")
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(logging.Formatter(_FILE_FORMAT, datefmt=_FILE_DATE_FORMAT))
        logger.addHandler(file_handler)
    except Exception as exc:
        _LOGGER.warning("Failed to configure file logging: %s", exc, exc_info=True)

    # Allow app/test harness handlers to capture logs.
    logger.propagate = True
    _logging_configured = True


def log_exception(context: str, exc: BaseException) -> Path | None:
    try:
        log_dir = get_log_dir()
        log_dir.mkdir(parents=True, exist_ok=True)
        path = get_log_path()
        timestamp = datetime.now().isoformat()
        with path.open("a", encoding="utf-8") as handle:
            handle.write(f"[{timestamp}] {context} failed: {type(exc).__name__}: {exc}\n")
            traceback.print_exception(type(exc), exc, exc.__traceback__, file=handle)
            handle.write("\n")
        return path
    except Exception as log_exc:
        _LOGGER.warning("Failed to write log file: %s", log_exc, exc_info=True)
        return None



from __future__ import annotations

import argparse
import logging
import os
from pathlib import Path
from typing import Iterable

from rich.console import Console

from .audio import SAMPLE_RATE
from .dx import render
from .errors import ModelNotAvailableError
from .logging_utils import configure_logging, log_exception
from .spinner import Spinner, render_error

_EXPRESSIVE_REPO = "mlx-community/gemma-3-1b-it-qat-8bit"
_EXPRESSIVE_DIR = "gemma-3-1b-it-qat-8bit"
_LOGGER = logging.getLogger("latentscore.cli")
_CONSOLE = Console()


def _default_model_base() -> Path:
    configured = os.environ.get("LATENTSCORE_MODEL_DIR")
    if configured:
        return Path(configured)
    return Path.home() / ".cache" / "latentscore" / "models"


def _download_expressive(model_base: Path) -> Path:
    try:
        from huggingface_hub import snapshot_download  # type: ignore[import]
    except ImportError as exc:
        _LOGGER.warning("huggingface_hub not installed: %s", exc, exc_info=True)
        raise ModelNotAvailableError("huggingface_hub is not installed") from exc

    target = model_base / _EXPRESSIVE_DIR
    target.mkdir(parents=True, exist_ok=True)
    snapshot_download(_EXPRESSIVE_REPO, local_dir=str(target))
    return target


def _doctor_report(lines: Iterable[str]) -> None:
    for line in lines:
        _CONSOLE.print(line)


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="latentscore")
    sub = parser.add_subparsers(dest="command", required=True)

    demo = sub.add_parser("demo", help="Render a short demo clip.")
    demo.add_argument("--duration", type=float, default=2.5)
    demo.add_argument("--output", type=str, default="demo.wav")

    download = sub.add_parser("download", help="Download model assets.")
    download.add_argument("model", choices=["expressive"], type=str)

    sub.add_parser("doctor", help="Check model availability and cache paths.")
    return parser


def main(argv: list[str] | None = None) -> int:
    configure_logging()
    try:
        parser = build_parser()
        args = parser.parse_args(argv)

        if args.command == "demo":
            with Spinner("Rendering demo audio"):
                audio = render("warm sunrise", duration=args.duration)
            path = audio.save(args.output)
            _CONSOLE.print(f"Wrote demo to {path} (sr={SAMPLE_RATE})")
            return 0

        if args.command == "download":
            model_base = _default_model_base()
            model_base.mkdir(parents=True, exist_ok=True)
            if args.model == "expressive":
                target = model_base / _EXPRESSIVE_DIR
                if not target.exists():
                    with Spinner("Downloading expressive model"):
                        _download_expressive(model_base)
                _CONSOLE.print(f"Downloaded expressive model to {target}")
                return 0

        if args.command == "doctor":
            model_base = _default_model_base()
            expressive_dir = model_base / _EXPRESSIVE_DIR
            embeddings_dir = Path("models") / "all-MiniLM-L6-v2"
            report = [
                f"Model cache base: {model_base}",
                f"Expressive model present: {expressive_dir.exists()}",
                f"Embeddings model present: {embeddings_dir.exists()}",
                "Hints:",
                "- Run `latentscore download expressive` to prefetch the LLM weights.",
                "- Set LATENTSCORE_MODEL_DIR to point at a preseeded models directory.",
                "- In production, run `latentscore doctor` and prefetch missing models to avoid runtime downloads.",
            ]
            _doctor_report(report)
            return 0

        parser.print_help()
        return 1
    except Exception as exc:
        debug = bool(os.environ.get("LATENTSCORE_DEBUG"))
        _LOGGER.warning("latentscore CLI failed: %s", exc, exc_info=debug)
        log_exception("latentscore CLI", exc)
        render_error("latentscore CLI", exc)
        return 1


if __name__ == "__main__":
    raise SystemExit(main())



from latentscore.models import ExternalModelSpec, resolve_model
from latentscore.providers.litellm import LiteLLMAdapter


def test_resolve_model_caches_builtin_instances() -> None:
    first = resolve_model("fast")
    second = resolve_model("fast")
    assert first is second


def test_resolve_model_external_prefix() -> None:
    model = resolve_model("external:gemini/gemini-3-flash-preview")
    assert isinstance(model, LiteLLMAdapter)


def test_resolve_model_external_spec() -> None:
    spec = ExternalModelSpec(model="gemini/gemini-3-flash-preview", api_key=None)
    model = resolve_model(spec)
    assert isinstance(model, LiteLLMAdapter)



from __future__ import annotations

import sys
from typing import IO

from .main import ModelLoadRole, RenderHooks, Streamable, StreamEvent, StreamHooks
from .models import EXTERNAL_PREFIX, ExternalModelSpec, ModelSpec
from .spinner import Spinner, render_error


class RichIndicator:
    """Render/stream indicator that maps hooks to spinner updates."""

    def __init__(
        self,
        *,
        stream: IO[str] | None = None,
        enabled: bool | None = None,
    ) -> None:
        self._stream = stream or sys.stderr
        self._spinner = Spinner("Starting", stream=self._stream, enabled=enabled)
        self._started = False
        self._stopped = False
        self._preview_items: set[int] = set()

    def render_hooks(self) -> RenderHooks:
        return RenderHooks(
            on_start=self._on_render_start,
            on_model_start=self._on_render_model_start,
            on_model_end=self._on_render_model_end,
            on_synth_start=self._on_render_synth_start,
            on_synth_end=self._on_render_synth_end,
            on_end=self._on_render_end,
            on_error=self._on_render_error,
        )

    def stream_hooks(self) -> StreamHooks:
        return StreamHooks(
            on_event=self._on_stream_event,
            on_model_load_start=self._on_stream_model_load_start,
            on_model_load_end=self._on_stream_model_load_end,
            on_stream_start=self._on_stream_start,
            on_item_resolve_start=self._on_item_resolve_start,
            on_item_resolve_success=self._on_item_resolve_success,
            on_item_resolve_error=self._on_item_resolve_error,
            on_item_preview_start=self._on_item_preview_start,
            on_first_config_ready=self._on_first_config_ready,
            on_first_audio_chunk=self._on_first_audio_chunk,
            on_stream_end=self._on_stream_end,
            on_error=self._on_stream_error,
        )

    def _start(self, message: str) -> None:
        if self._stopped:
            return
        if not self._started:
            self._spinner.update(message)
            self._spinner.start()
            self._started = True
            return
        self._spinner.update(message)

    def _stop(self) -> None:
        if self._stopped:
            return
        self._stopped = True
        self._spinner.stop()

    def _on_render_start(self) -> None:
        self._start("Preparing render")

    def _on_render_model_start(self, model: ModelSpec) -> None:
        if _is_external_model(model):
            self._start("Generating config (LLM)")
        else:
            self._start("Loading model")

    def _on_render_model_end(self, model: ModelSpec) -> None:
        _ = model
        self._start("Generating music")

    def _on_render_synth_start(self) -> None:
        self._start("Generating music")

    def _on_render_synth_end(self) -> None:
        self._start("Finalizing audio")

    def _on_render_end(self) -> None:
        self._stop()

    def _on_render_error(self, exc: Exception) -> None:
        self._stop()
        render_error("render", exc, stream=self._stream)

    def _on_stream_event(self, event: StreamEvent) -> None:
        _ = event

    def _on_stream_model_load_start(self, model: ModelSpec, role: ModelLoadRole) -> None:
        _ = role
        if _is_external_model(model):
            self._start("Preparing LLM session")
        else:
            self._start("Loading model")

    def _on_stream_model_load_end(self, model: ModelSpec, role: ModelLoadRole) -> None:
        _ = model
        _ = role
        if not self._started:
            self._start("Preparing stream")

    def _on_stream_start(self) -> None:
        self._start("Preparing stream")

    def _on_item_resolve_start(self, index: int, item: Streamable) -> None:
        label = _item_action(item)
        self._start(f"{label} (item {index + 1})")

    def _on_item_preview_start(
        self,
        index: int,
        item: Streamable,
        preview: bool,
    ) -> None:
        _ = item
        self._preview_items.add(index)
        label = "speculative" if preview else "standard"
        self._start(f"Generating preview ({label}) for item {index + 1}")

    def _on_item_resolve_success(self, index: int, item: Streamable) -> None:
        _ = item
        if index in self._preview_items:
            self._preview_items.discard(index)
            self._start(f"Applying final config for item {index + 1}")

    def _on_item_resolve_error(
        self,
        index: int,
        item: Streamable,
        exc: Exception,
        fallback: object | None,
    ) -> None:
        _ = index
        _ = item
        _ = exc
        _ = fallback
        self._start("Config failed; applying fallback")

    def _on_first_config_ready(self, index: int, item: Streamable) -> None:
        _ = index
        _ = item
        self._start("Generating music")

    def _on_first_audio_chunk(self) -> None:
        self._start("Streaming audio")

    def _on_stream_end(self) -> None:
        self._stop()

    def _on_stream_error(self, exc: Exception) -> None:
        self._stop()
        render_error("stream", exc, stream=self._stream)


def _is_external_model(model: ModelSpec) -> bool:
    if isinstance(model, ExternalModelSpec):
        return True
    if isinstance(model, str):
        return model.startswith(EXTERNAL_PREFIX)
    return False


def _item_action(item: Streamable) -> str:
    if isinstance(item.content, str):
        return "Generating config (LLM)"
    return "Applying config"



from __future__ import annotations

import logging
from types import MappingProxyType
from typing import Any, Callable, Literal, Mapping, Optional, TypeVar, cast

from pydantic import BaseModel, ConfigDict, Field, ValidationError, model_validator

from .errors import InvalidConfigError

_LOGGER = logging.getLogger("latentscore.config")

TempoLabel = Literal["very_slow", "slow", "medium", "fast", "very_fast"]
BrightnessLabel = Literal["very_dark", "dark", "medium", "bright", "very_bright"]
SpaceLabel = Literal["dry", "small", "medium", "large", "vast"]
MotionLabel = Literal["static", "slow", "medium", "fast", "chaotic"]
StereoLabel = Literal["mono", "narrow", "medium", "wide", "ultra_wide"]
EchoLabel = Literal["none", "subtle", "medium", "heavy", "infinite"]
HumanFeelLabel = Literal["robotic", "tight", "natural", "loose", "drunk"]
RootNote = Literal["c", "c#", "d", "d#", "e", "f", "f#", "g", "g#", "a", "a#", "b"]
ModeName = Literal["major", "minor", "dorian", "mixolydian"]
DensityLevel = Literal[2, 3, 4, 5, 6]
BassStyle = Literal[
    "drone",
    "sustained",
    "pulsing",
    "walking",
    "fifth_drone",
    "sub_pulse",
    "octave",
    "arp_bass",
]
PadStyle = Literal[
    "warm_slow",
    "dark_sustained",
    "cinematic",
    "thin_high",
    "ambient_drift",
    "stacked_fifths",
    "bright_open",
]
MelodyStyle = Literal[
    "procedural",
    "contemplative",
    "rising",
    "falling",
    "minimal",
    "ornamental",
    "arp_melody",
    "contemplative_minor",
    "call_response",
    "heroic",
]
RhythmStyle = Literal[
    "none",
    "minimal",
    "heartbeat",
    "soft_four",
    "hats_only",
    "electronic",
    "kit_light",
    "kit_medium",
    "military",
    "tabla_essence",
    "brush",
]
TextureStyle = Literal[
    "none",
    "shimmer",
    "shimmer_slow",
    "vinyl_crackle",
    "breath",
    "stars",
    "glitch",
    "noise_wash",
    "crystal",
    "pad_whisper",
]
AccentStyle = Literal[
    "none",
    "bells",
    "pluck",
    "chime",
    "bells_dense",
    "blip",
    "blip_random",
    "brass_hit",
    "wind",
    "arp_accent",
    "piano_note",
]
AttackStyle = Literal["soft", "medium", "sharp"]
GrainStyle = Literal["clean", "warm", "gritty"]
MelodyEngine = Literal["pattern", "procedural"]
TensionCurve = Literal["arc", "ramp", "waves"]
HarmonyStyle = Literal["auto", "pop", "jazz", "cinematic", "ambient"]
ChordExtensions = Literal["triads", "sevenths", "lush"]
PhraseLengthBars = Literal[2, 4, 8]
MelodyDensityLabel = Literal["very_sparse", "sparse", "medium", "busy", "very_busy"]
SyncopationLabel = Literal["straight", "light", "medium", "heavy"]
SwingLabel = Literal["none", "light", "medium", "heavy"]
MotifRepeatLabel = Literal["rare", "sometimes", "often"]
StepBiasLabel = Literal["step", "balanced", "leapy"]
ChromaticLabel = Literal["none", "light", "medium", "heavy"]
CadenceLabel = Literal["weak", "medium", "strong"]
ChordChangeLabel = Literal["very_slow", "slow", "medium", "fast"]


T = TypeVar("T")

_TEMPO_MAP: Mapping[TempoLabel, float] = MappingProxyType(
    {
        "very_slow": 0.15,
        "slow": 0.3,
        "medium": 0.5,
        "fast": 0.7,
        "very_fast": 0.9,
    }
)
_BRIGHTNESS_MAP: Mapping[BrightnessLabel, float] = MappingProxyType(
    {
        "very_dark": 0.1,
        "dark": 0.3,
        "medium": 0.5,
        "bright": 0.7,
        "very_bright": 0.9,
    }
)
_SPACE_MAP: Mapping[SpaceLabel, float] = MappingProxyType(
    {
        "dry": 0.1,
        "small": 0.3,
        "medium": 0.5,
        "large": 0.7,
        "vast": 0.95,
    }
)
_MOTION_MAP: Mapping[MotionLabel, float] = MappingProxyType(
    {
        "static": 0.1,
        "slow": 0.3,
        "medium": 0.5,
        "fast": 0.7,
        "chaotic": 0.9,
    }
)
_STEREO_MAP: Mapping[StereoLabel, float] = MappingProxyType(
    {
        "mono": 0.0,
        "narrow": 0.25,
        "medium": 0.5,
        "wide": 0.75,
        "ultra_wide": 1.0,
    }
)
_ECHO_MAP: Mapping[EchoLabel, float] = MappingProxyType(
    {
        "none": 0.0,
        "subtle": 0.25,
        "medium": 0.5,
        "heavy": 0.75,
        "infinite": 0.95,
    }
)
_HUMAN_MAP: Mapping[HumanFeelLabel, float] = MappingProxyType(
    {
        "robotic": 0.0,
        "tight": 0.15,
        "natural": 0.3,
        "loose": 0.5,
        "drunk": 0.8,
    }
)
_MELODY_DENSITY_MAP: Mapping[MelodyDensityLabel, float] = MappingProxyType(
    {
        "very_sparse": 0.15,
        "sparse": 0.30,
        "medium": 0.50,
        "busy": 0.70,
        "very_busy": 0.85,
    }
)
_SYNCOPATION_MAP: Mapping[SyncopationLabel, float] = MappingProxyType(
    {
        "straight": 0.0,
        "light": 0.2,
        "medium": 0.5,
        "heavy": 0.8,
    }
)
_SWING_MAP: Mapping[SwingLabel, float] = MappingProxyType(
    {
        "none": 0.0,
        "light": 0.2,
        "medium": 0.5,
        "heavy": 0.8,
    }
)
_MOTIF_REPEAT_MAP: Mapping[MotifRepeatLabel, float] = MappingProxyType(
    {
        "rare": 0.2,
        "sometimes": 0.5,
        "often": 0.8,
    }
)
_STEP_BIAS_MAP: Mapping[StepBiasLabel, float] = MappingProxyType(
    {
        "step": 0.9,
        "balanced": 0.7,
        "leapy": 0.4,
    }
)
_CHROMATIC_MAP: Mapping[ChromaticLabel, float] = MappingProxyType(
    {
        "none": 0.0,
        "light": 0.05,
        "medium": 0.12,
        "heavy": 0.25,
    }
)
_CADENCE_MAP: Mapping[CadenceLabel, float] = MappingProxyType(
    {
        "weak": 0.3,
        "medium": 0.6,
        "strong": 0.9,
    }
)
_CHORD_CHANGE_BARS_MAP: Mapping[ChordChangeLabel, int] = MappingProxyType(
    {
        "very_slow": 4,
        "slow": 2,
        "medium": 1,
        "fast": 1,
    }
)


def _optional_map(value: T | None, mapper: Callable[[T], float]) -> float | None:
    match value:
        case None:
            return None
        case _:
            return mapper(value)


def tempo_to_float(value: TempoLabel) -> float:
    try:
        return _TEMPO_MAP[value]
    except KeyError as exc:
        raise InvalidConfigError(f"Unknown tempo label: {value!r}") from exc


def brightness_to_float(value: BrightnessLabel) -> float:
    try:
        return _BRIGHTNESS_MAP[value]
    except KeyError as exc:
        raise InvalidConfigError(f"Unknown brightness label: {value!r}") from exc


def space_to_float(value: SpaceLabel) -> float:
    try:
        return _SPACE_MAP[value]
    except KeyError as exc:
        raise InvalidConfigError(f"Unknown space label: {value!r}") from exc


def motion_to_float(value: MotionLabel) -> float:
    try:
        return _MOTION_MAP[value]
    except KeyError as exc:
        raise InvalidConfigError(f"Unknown motion label: {value!r}") from exc


def stereo_to_float(value: StereoLabel) -> float:
    try:
        return _STEREO_MAP[value]
    except KeyError as exc:
        raise InvalidConfigError(f"Unknown stereo label: {value!r}") from exc


def echo_to_float(value: EchoLabel) -> float:
    try:
        return _ECHO_MAP[value]
    except KeyError as exc:
        raise InvalidConfigError(f"Unknown echo label: {value!r}") from exc


def human_to_float(value: HumanFeelLabel) -> float:
    try:
        return _HUMAN_MAP[value]
    except KeyError as exc:
        raise InvalidConfigError(f"Unknown human feel label: {value!r}") from exc


def melody_density_to_float(value: MelodyDensityLabel) -> float:
    try:
        return _MELODY_DENSITY_MAP[value]
    except KeyError as exc:
        raise InvalidConfigError(f"Unknown melody density label: {value!r}") from exc


def syncopation_to_float(value: SyncopationLabel) -> float:
    try:
        return _SYNCOPATION_MAP[value]
    except KeyError as exc:
        raise InvalidConfigError(f"Unknown syncopation label: {value!r}") from exc


def swing_to_float(value: SwingLabel) -> float:
    try:
        return _SWING_MAP[value]
    except KeyError as exc:
        raise InvalidConfigError(f"Unknown swing label: {value!r}") from exc


def motif_repeat_to_float(value: MotifRepeatLabel) -> float:
    try:
        return _MOTIF_REPEAT_MAP[value]
    except KeyError as exc:
        raise InvalidConfigError(f"Unknown motif repeat label: {value!r}") from exc


def step_bias_to_float(value: StepBiasLabel) -> float:
    try:
        return _STEP_BIAS_MAP[value]
    except KeyError as exc:
        raise InvalidConfigError(f"Unknown step bias label: {value!r}") from exc


def chromatic_to_float(value: ChromaticLabel) -> float:
    try:
        return _CHROMATIC_MAP[value]
    except KeyError as exc:
        raise InvalidConfigError(f"Unknown chromatic label: {value!r}") from exc


def cadence_to_float(value: CadenceLabel) -> float:
    try:
        return _CADENCE_MAP[value]
    except KeyError as exc:
        raise InvalidConfigError(f"Unknown cadence label: {value!r}") from exc


def chord_change_to_bars(value: ChordChangeLabel) -> int:
    try:
        return _CHORD_CHANGE_BARS_MAP[value]
    except KeyError as exc:
        raise InvalidConfigError(f"Unknown chord change label: {value!r}") from exc


class _MusicConfigInternal(BaseModel):
    """Internal config with numeric values used by the synth engine."""

    tempo: float = 0.35
    root: RootNote = "c"
    mode: ModeName = "minor"
    brightness: float = 0.5
    space: float = 0.6
    density: DensityLevel = 5

    bass: BassStyle = "drone"
    pad: PadStyle = "warm_slow"
    melody: MelodyStyle = "contemplative"
    rhythm: RhythmStyle = "minimal"
    texture: TextureStyle = "shimmer"
    accent: AccentStyle = "bells"

    motion: float = 0.5
    attack: AttackStyle = "medium"
    stereo: float = 0.5
    depth: bool = False
    echo: float = 0.5
    human: float = 0.0
    grain: GrainStyle = "clean"

    melody_engine: MelodyEngine = "pattern"
    phrase_len_bars: int = 4
    melody_density: float = 0.45
    syncopation: float = 0.20
    swing: float = 0.0
    motif_repeat_prob: float = 0.50
    step_bias: float = 0.75
    chromatic_prob: float = 0.05
    cadence_strength: float = 0.65
    register_min_oct: int = 4
    register_max_oct: int = 6
    tension_curve: TensionCurve = "arc"
    harmony_style: HarmonyStyle = "auto"
    chord_change_bars: int = 1
    chord_extensions: ChordExtensions = "triads"

    model_config = ConfigDict(extra="ignore")

    @model_validator(mode="before")
    @classmethod
    def _flatten_layers(cls, data: object) -> object:
        if not isinstance(data, Mapping):
            return data
        merged: dict[str, Any] = dict(cast(Mapping[str, Any], data))
        layers = merged.pop("layers", None)
        if isinstance(layers, Mapping):
            for key, value in cast(Mapping[str, Any], layers).items():
                merged.setdefault(key, value)
        return merged

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> "_MusicConfigInternal":
        """Create config from dict (e.g., from JSON)."""
        return cls.model_validate(data)


class MusicConfig(BaseModel):
    """Strictly-typed public config with tolerant extra capture."""

    schema_version: Literal[1] = 1
    tempo: TempoLabel = "medium"
    root: RootNote = "c"
    mode: ModeName = "minor"
    brightness: BrightnessLabel = "medium"
    space: SpaceLabel = "medium"
    density: DensityLevel = 5

    bass: BassStyle = "drone"
    pad: PadStyle = "warm_slow"
    melody: MelodyStyle = "contemplative"
    rhythm: RhythmStyle = "minimal"
    texture: TextureStyle = "shimmer"
    accent: AccentStyle = "bells"

    motion: MotionLabel = "medium"
    attack: AttackStyle = "medium"
    stereo: StereoLabel = "medium"
    depth: bool = False
    echo: EchoLabel = "medium"
    human: HumanFeelLabel = "robotic"
    grain: GrainStyle = "clean"

    melody_engine: MelodyEngine = "pattern"
    phrase_len_bars: int = 4
    melody_density: float = 0.45
    syncopation: float = 0.20
    swing: float = 0.0
    motif_repeat_prob: float = 0.50
    step_bias: float = 0.75
    chromatic_prob: float = 0.05
    cadence_strength: float = 0.65
    register_min_oct: int = 4
    register_max_oct: int = 6
    tension_curve: TensionCurve = "arc"
    harmony_style: HarmonyStyle = "auto"
    chord_change_bars: int = 1
    chord_extensions: ChordExtensions = "triads"

    model_config = ConfigDict(extra="allow")

    @property
    def extras(self) -> dict[str, Any]:
        return dict(self.model_extra or {})

    def to_internal(self) -> _MusicConfigInternal:
        return _MusicConfigInternal(
            tempo=tempo_to_float(self.tempo),
            brightness=brightness_to_float(self.brightness),
            root=self.root,
            mode=self.mode,
            space=space_to_float(self.space),
            density=self.density,
            bass=self.bass,
            pad=self.pad,
            melody=self.melody,
            rhythm=self.rhythm,
            texture=self.texture,
            accent=self.accent,
            motion=motion_to_float(self.motion),
            attack=self.attack,
            stereo=stereo_to_float(self.stereo),
            depth=self.depth,
            echo=echo_to_float(self.echo),
            human=human_to_float(self.human),
            grain=self.grain,
            melody_engine=self.melody_engine,
            phrase_len_bars=self.phrase_len_bars,
            melody_density=self.melody_density,
            syncopation=self.syncopation,
            swing=self.swing,
            motif_repeat_prob=self.motif_repeat_prob,
            step_bias=self.step_bias,
            chromatic_prob=self.chromatic_prob,
            cadence_strength=self.cadence_strength,
            register_min_oct=self.register_min_oct,
            register_max_oct=self.register_max_oct,
            tension_curve=self.tension_curve,
            harmony_style=self.harmony_style,
            chord_change_bars=self.chord_change_bars,
            chord_extensions=self.chord_extensions,
        )


SynthConfig = _MusicConfigInternal


_PROMPT_DESC: dict[str, str] = {
    "justification": (
        "Explain the sonic reasoning for the choices. Mention vibe decomposition, sonic "
        "translation, coherence check, and which examples guided the selection."
    ),
    "config": "Music configuration that matches the requested vibe.",
    "tempo": "Tempo label controlling overall speed and energy.",
    "root": "Root note of the scale.",
    "mode": "Scale mode that shapes the emotional color.",
    "brightness": "Filter brightness / spectral tilt label.",
    "space": "Reverb/room size label.",
    "density": "Layer count indicating overall thickness.",
    "bass": "Bass style or movement pattern.",
    "pad": "Pad texture and harmonic bed style.",
    "melody": "Melody style or contour.",
    "rhythm": "Percussion pattern style (or none).",
    "texture": "Background texture or noise layer.",
    "accent": "Sparse accent sound type.",
    "motion": "Modulation/LFO rate label.",
    "attack": "Transient sharpness label.",
    "stereo": "Stereo width label.",
    "depth": "Whether to add sub-bass depth.",
    "echo": "Delay amount label.",
    "human": "Timing/pitch looseness label.",
    "grain": "Oscillator character (clean/warm/gritty).",
    "melody_engine": "Melody generation mode (procedural or pattern).",
    "phrase_len_bars": "Phrase length in bars (2, 4, or 8).",
    "melody_density": "Melody note density label (very_sparse, sparse, medium, busy, very_busy).",
    "syncopation": "Offbeat emphasis label (straight, light, medium, heavy).",
    "swing": "Swing amount label (none, light, medium, heavy).",
    "motif_repeat_prob": "Motif repetition label (rare, sometimes, often).",
    "step_bias": "Melodic motion label (step, balanced, leapy).",
    "chromatic_prob": "Chromaticism label (none, light, medium, heavy).",
    "cadence_strength": "Cadence emphasis label (weak, medium, strong).",
    "register_min_oct": "Lowest melody octave (integer).",
    "register_max_oct": "Highest melody octave (integer).",
    "tension_curve": "Tension shape across the phrase.",
    "harmony_style": "Harmony progression style.",
    "chord_change_bars": "Chord change rate label (very_slow, slow, medium, fast).",
    "chord_extensions": "Chord color/extension level.",
}

_PROMPT_REGISTER_MIN = 1
_PROMPT_REGISTER_MAX = 8


class MusicConfigPrompt(BaseModel):
    """Prompt-only schema matching MusicConfig without defaults."""

    tempo: TempoLabel = Field(description=_PROMPT_DESC["tempo"])
    root: RootNote = Field(description=_PROMPT_DESC["root"])
    mode: ModeName = Field(description=_PROMPT_DESC["mode"])
    brightness: BrightnessLabel = Field(description=_PROMPT_DESC["brightness"])
    space: SpaceLabel = Field(description=_PROMPT_DESC["space"])
    density: DensityLevel = Field(ge=2, le=6, description=_PROMPT_DESC["density"])

    bass: BassStyle = Field(description=_PROMPT_DESC["bass"])
    pad: PadStyle = Field(description=_PROMPT_DESC["pad"])
    melody: MelodyStyle = Field(description=_PROMPT_DESC["melody"])
    rhythm: RhythmStyle = Field(description=_PROMPT_DESC["rhythm"])
    texture: TextureStyle = Field(description=_PROMPT_DESC["texture"])
    accent: AccentStyle = Field(description=_PROMPT_DESC["accent"])

    motion: MotionLabel = Field(description=_PROMPT_DESC["motion"])
    attack: AttackStyle = Field(description=_PROMPT_DESC["attack"])
    stereo: StereoLabel = Field(description=_PROMPT_DESC["stereo"])
    depth: bool = Field(description=_PROMPT_DESC["depth"])
    echo: EchoLabel = Field(description=_PROMPT_DESC["echo"])
    human: HumanFeelLabel = Field(description=_PROMPT_DESC["human"])
    grain: GrainStyle = Field(description=_PROMPT_DESC["grain"])

    melody_engine: MelodyEngine = Field(description=_PROMPT_DESC["melody_engine"])
    phrase_len_bars: PhraseLengthBars = Field(description=_PROMPT_DESC["phrase_len_bars"])
    melody_density: MelodyDensityLabel = Field(description=_PROMPT_DESC["melody_density"])
    syncopation: SyncopationLabel = Field(description=_PROMPT_DESC["syncopation"])
    swing: SwingLabel = Field(description=_PROMPT_DESC["swing"])
    motif_repeat_prob: MotifRepeatLabel = Field(description=_PROMPT_DESC["motif_repeat_prob"])
    step_bias: StepBiasLabel = Field(description=_PROMPT_DESC["step_bias"])
    chromatic_prob: ChromaticLabel = Field(description=_PROMPT_DESC["chromatic_prob"])
    cadence_strength: CadenceLabel = Field(description=_PROMPT_DESC["cadence_strength"])
    register_min_oct: int = Field(
        ge=_PROMPT_REGISTER_MIN,
        le=_PROMPT_REGISTER_MAX,
        description=_PROMPT_DESC["register_min_oct"],
    )
    register_max_oct: int = Field(
        ge=_PROMPT_REGISTER_MIN,
        le=_PROMPT_REGISTER_MAX,
        description=_PROMPT_DESC["register_max_oct"],
    )
    tension_curve: TensionCurve = Field(description=_PROMPT_DESC["tension_curve"])
    harmony_style: HarmonyStyle = Field(description=_PROMPT_DESC["harmony_style"])
    chord_change_bars: ChordChangeLabel = Field(description=_PROMPT_DESC["chord_change_bars"])
    chord_extensions: ChordExtensions = Field(description=_PROMPT_DESC["chord_extensions"])

    def to_config(self) -> MusicConfig:
        data = self.model_dump()
        data["melody_density"] = melody_density_to_float(self.melody_density)
        data["syncopation"] = syncopation_to_float(self.syncopation)
        data["swing"] = swing_to_float(self.swing)
        data["motif_repeat_prob"] = motif_repeat_to_float(self.motif_repeat_prob)
        data["step_bias"] = step_bias_to_float(self.step_bias)
        data["chromatic_prob"] = chromatic_to_float(self.chromatic_prob)
        data["cadence_strength"] = cadence_to_float(self.cadence_strength)
        data["chord_change_bars"] = chord_change_to_bars(self.chord_change_bars)
        return MusicConfig.model_validate(data)

    model_config = ConfigDict(extra="forbid")


class MusicConfigPromptPayload(BaseModel):
    """LLM payload that includes a justification and a config."""

    justification: str = Field(description=_PROMPT_DESC["justification"])
    config: MusicConfigPrompt = Field(description=_PROMPT_DESC["config"])

    model_config = ConfigDict(extra="forbid")


def _assert_prompt_schema_parity() -> None:
    prompt_fields = set(MusicConfigPrompt.model_fields)
    config_fields = set(MusicConfig.model_fields)
    excluded = {"schema_version"}
    if prompt_fields == (config_fields - excluded):
        return
    missing = sorted(config_fields - prompt_fields - excluded)
    extra = sorted(prompt_fields - config_fields)
    raise AssertionError(f"MusicConfigPrompt mismatch: missing={missing!r}, extra={extra!r}")


_assert_prompt_schema_parity()


class _MusicConfigUpdateInternal(BaseModel):
    """Partial update for internal numeric configs."""

    tempo: Optional[float] = None
    brightness: Optional[float] = None
    root: Optional[RootNote] = None
    mode: Optional[ModeName] = None
    space: Optional[float] = None
    density: Optional[DensityLevel] = None

    bass: Optional[BassStyle] = None
    pad: Optional[PadStyle] = None
    melody: Optional[MelodyStyle] = None
    rhythm: Optional[RhythmStyle] = None
    texture: Optional[TextureStyle] = None
    accent: Optional[AccentStyle] = None

    motion: Optional[float] = None
    attack: Optional[AttackStyle] = None
    stereo: Optional[float] = None
    depth: Optional[bool] = None
    echo: Optional[float] = None
    human: Optional[float] = None
    grain: Optional[GrainStyle] = None

    melody_engine: Optional[MelodyEngine] = None
    phrase_len_bars: Optional[int] = None
    melody_density: Optional[float] = None
    syncopation: Optional[float] = None
    swing: Optional[float] = None
    motif_repeat_prob: Optional[float] = None
    step_bias: Optional[float] = None
    chromatic_prob: Optional[float] = None
    cadence_strength: Optional[float] = None
    register_min_oct: Optional[int] = None
    register_max_oct: Optional[int] = None
    tension_curve: Optional[TensionCurve] = None
    harmony_style: Optional[HarmonyStyle] = None
    chord_change_bars: Optional[int] = None
    chord_extensions: Optional[ChordExtensions] = None

    model_config = ConfigDict(extra="forbid")


class MusicConfigUpdate(BaseModel):
    """Partial update for human inputs; unknown keys are rejected."""

    tempo: Optional[TempoLabel] = None
    brightness: Optional[BrightnessLabel] = None
    root: Optional[RootNote] = None
    mode: Optional[ModeName] = None
    space: Optional[SpaceLabel] = None
    density: Optional[DensityLevel] = None

    bass: Optional[BassStyle] = None
    pad: Optional[PadStyle] = None
    melody: Optional[MelodyStyle] = None
    rhythm: Optional[RhythmStyle] = None
    texture: Optional[TextureStyle] = None
    accent: Optional[AccentStyle] = None

    motion: Optional[MotionLabel] = None
    attack: Optional[AttackStyle] = None
    stereo: Optional[StereoLabel] = None
    depth: Optional[bool] = None
    echo: Optional[EchoLabel] = None
    human: Optional[HumanFeelLabel] = None
    grain: Optional[GrainStyle] = None

    melody_engine: Optional[MelodyEngine] = None
    phrase_len_bars: Optional[int] = None
    melody_density: Optional[float] = None
    syncopation: Optional[float] = None
    swing: Optional[float] = None
    motif_repeat_prob: Optional[float] = None
    step_bias: Optional[float] = None
    chromatic_prob: Optional[float] = None
    cadence_strength: Optional[float] = None
    register_min_oct: Optional[int] = None
    register_max_oct: Optional[int] = None
    tension_curve: Optional[TensionCurve] = None
    harmony_style: Optional[HarmonyStyle] = None
    chord_change_bars: Optional[int] = None
    chord_extensions: Optional[ChordExtensions] = None

    model_config = ConfigDict(extra="forbid")

    def to_internal(self) -> _MusicConfigUpdateInternal:
        return _MusicConfigUpdateInternal(
            tempo=_optional_map(self.tempo, tempo_to_float),
            brightness=_optional_map(self.brightness, brightness_to_float),
            root=self.root,
            mode=self.mode,
            space=_optional_map(self.space, space_to_float),
            density=self.density,
            bass=self.bass,
            pad=self.pad,
            melody=self.melody,
            rhythm=self.rhythm,
            texture=self.texture,
            accent=self.accent,
            motion=_optional_map(self.motion, motion_to_float),
            attack=self.attack,
            stereo=_optional_map(self.stereo, stereo_to_float),
            depth=self.depth,
            echo=_optional_map(self.echo, echo_to_float),
            human=_optional_map(self.human, human_to_float),
            grain=self.grain,
            melody_engine=self.melody_engine,
            phrase_len_bars=self.phrase_len_bars,
            melody_density=self.melody_density,
            syncopation=self.syncopation,
            swing=self.swing,
            motif_repeat_prob=self.motif_repeat_prob,
            step_bias=self.step_bias,
            chromatic_prob=self.chromatic_prob,
            cadence_strength=self.cadence_strength,
            register_min_oct=self.register_min_oct,
            register_max_oct=self.register_max_oct,
            tension_curve=self.tension_curve,
            harmony_style=self.harmony_style,
            chord_change_bars=self.chord_change_bars,
            chord_extensions=self.chord_extensions,
        )


ConfigInput = MusicConfig | _MusicConfigInternal
UpdateInput = MusicConfigUpdate | _MusicConfigUpdateInternal


def coerce_internal_config(config: ConfigInput) -> _MusicConfigInternal:
    match config:
        case _MusicConfigInternal():
            return config
        case MusicConfig():
            return config.to_internal()
        case _:
            raise InvalidConfigError(f"Unsupported config type: {type(config).__name__}")


def coerce_internal_update(update: UpdateInput) -> _MusicConfigUpdateInternal:
    match update:
        case _MusicConfigUpdateInternal():
            return update
        case MusicConfigUpdate():
            return update.to_internal()
        case _:
            raise InvalidConfigError(f"Unsupported update type: {type(update).__name__}")


def merge_config(base: MusicConfig, update: MusicConfigUpdate) -> MusicConfig:
    """Apply a partial update to a base config."""

    changes = update.model_dump(exclude_none=True)
    return base.model_copy(update=changes)


def merge_internal_config(
    base: _MusicConfigInternal,
    update: _MusicConfigUpdateInternal,
) -> _MusicConfigInternal:
    """Apply a partial update to a base internal config."""

    changes = update.model_dump(exclude_none=True)
    return base.model_copy(update=changes)


def parse_update(payload: Mapping[str, Any]) -> MusicConfigUpdate:
    """Parse a strict update payload, raising InvalidConfigError on failure."""

    try:
        return MusicConfigUpdate.model_validate(payload)
    except ValidationError as exc:  # pragma: no cover - defensive
        _LOGGER.warning("Failed to parse config update: %s", exc, exc_info=True)
        raise InvalidConfigError(str(exc)) from exc


def parse_config(payload: Mapping[str, Any]) -> MusicConfig:
    """Parse a tolerant config payload, raising InvalidConfigError on failure."""

    try:
        return MusicConfig.model_validate(payload)
    except ValidationError as exc:  # pragma: no cover - defensive
        _LOGGER.warning("Failed to parse config payload: %s", exc, exc_info=True)
        raise InvalidConfigError(str(exc)) from exc


def is_empty_update(update: UpdateInput) -> bool:
    """Return True when the update does not specify any fields."""

    return not update.model_dump(exclude_none=True)



from __future__ import annotations

from collections.abc import Iterable, Iterator, Sequence
from pathlib import Path
from typing import Any, Callable, TypeGuard

import numpy as np
import soundfile as sf  # type: ignore[import]
from numpy.typing import NDArray

from .errors import InvalidConfigError

FloatArray = NDArray[np.float32]
AudioNumbers = NDArray[np.floating[Any]] | Sequence[float] | FloatArray

SAMPLE_RATE = 44_100


def ensure_audio_contract(
    audio: AudioNumbers,
    *,
    sample_rate: int = SAMPLE_RATE,
    check_peak: bool = True,
) -> FloatArray:
    """Normalize dtype/range/shape to the audio contract (optionally skip peak checks)."""

    _ = sample_rate
    mono: FloatArray = np.asarray(audio, dtype=np.float32).reshape(-1)
    if mono.size == 0:
        return mono
    if check_peak:
        peak = float(np.max(np.abs(mono)))
        if peak > 1.0:
            mono = mono / peak
    return mono


def iter_chunks(chunks: Iterable[AudioNumbers]) -> Iterator[FloatArray]:
    """Yield chunks that already respect the audio contract."""

    for chunk in chunks:
        yield ensure_audio_contract(chunk)


def _is_audio_numbers(value: object) -> TypeGuard[AudioNumbers]:
    if isinstance(value, np.ndarray):
        return True
    if isinstance(value, Sequence) and not isinstance(value, (str, bytes)):
        return True
    return False


def _coerce_write_audio(
    write_fn: object,
) -> Callable[[Path | str, AudioNumbers, int], None]:
    if not callable(write_fn):
        raise InvalidConfigError("soundfile.write is unavailable")

    def _write_audio(path: Path | str, audio: AudioNumbers, sample_rate: int) -> None:
        write_fn(path, audio, sample_rate)

    return _write_audio


def _coerce_write_chunk(write_fn: object) -> Callable[[FloatArray], None]:
    if not callable(write_fn):
        raise InvalidConfigError("soundfile write handle is unavailable")

    def _write_chunk(chunk: FloatArray) -> None:
        write_fn(chunk)

    return _write_chunk


def _looks_like_samples(sequence: Sequence[object]) -> bool:
    match sequence:
        case []:
            return True
        case [int() | float() | np.floating(), *_]:
            return True
        case _:
            return False


def write_wav(
    path: str | Path,
    audio_or_chunks: AudioNumbers | Iterable[AudioNumbers],
    *,
    sample_rate: int = SAMPLE_RATE,
) -> Path:
    """Write a full array or chunk iterator to a wav file."""

    target = Path(path)
    audio_obj: object = audio_or_chunks
    match audio_obj:
        case np.ndarray():
            write_fn = getattr(sf, "write", None)
            write_audio = _coerce_write_audio(write_fn)
            array_float: NDArray[np.float32] = np.asarray(audio_obj, dtype=np.float32)
            normalized = ensure_audio_contract(array_float, sample_rate=sample_rate)
            # soundfile stubs are incomplete; keep runtime checks explicit.
            write_audio(target, normalized, sample_rate)  # type: ignore[reportUnknownMemberType]
            return target
        case Sequence() as sequence if _looks_like_samples(sequence):
            write_fn = getattr(sf, "write", None)
            write_audio = _coerce_write_audio(write_fn)
            assert _looks_like_samples(sequence)
            sequence_array: NDArray[np.float32] = np.asarray(sequence, dtype=np.float32)
            normalized = ensure_audio_contract(sequence_array, sample_rate=sample_rate)
            write_audio(target, normalized, sample_rate)  # type: ignore[reportUnknownMemberType]
            return target
        case str() | bytes():  # type: ignore[reportUnnecessaryComparison]
            raise InvalidConfigError("audio_or_chunks must be audio samples or chunk iterables")
        case Iterable() as chunks:
            match chunks:
                case str() | bytes():  # type: ignore[reportUnnecessaryComparison]
                    raise InvalidConfigError(
                        "audio_or_chunks must be audio samples or chunk iterables"
                    )
                case _:
                    pass
        case _:
            raise InvalidConfigError("audio_or_chunks must be audio samples or chunk iterables")

    with sf.SoundFile(
        target,
        mode="w",
        samplerate=sample_rate,
        channels=1,
        subtype="FLOAT",
    ) as handle:
        write_fn = getattr(handle, "write", None)
        write_chunk = _coerce_write_chunk(write_fn)  # type: ignore[reportUnknownMemberType]
        assert isinstance(chunks, Iterable)
        for chunk in chunks:
            if not _is_audio_numbers(chunk):
                raise InvalidConfigError("audio_or_chunks must be audio samples or chunk iterables")
            write_chunk(ensure_audio_contract(chunk, sample_rate=sample_rate))

    return target



from pathlib import Path

import numpy as np

from latentscore.audio import ensure_audio_contract, write_wav


def test_write_wav_accepts_sequence(tmp_path: Path) -> None:
    target = tmp_path / "seq.wav"
    samples = [0.0, 0.1, -0.1, 0.0]

    write_wav(target, samples, sample_rate=22_050)

    assert target.exists()
    assert target.stat().st_size > 0


def test_ensure_audio_contract_skip_peak() -> None:
    audio = np.array([2.0, -2.0], dtype=np.float32)
    out = ensure_audio_contract(audio, check_peak=False)
    assert np.allclose(out, audio)
